from typing import Mapping, Tuple

import aesara
import aesara.tensor as at
from aesara.tensor.random import RandomStream
from aesara.tensor.var import TensorVariable


def ffbs_step(
    gamma_0: TensorVariable,
    Gammas: TensorVariable,
    log_lik: TensorVariable,
    srng: RandomStream,
    lower_prec_bound: float = 1e-20,
) -> Tuple[TensorVariable, Mapping]:
    r"""Draw a discrete state sequence sample using forward-filtering backward-sampling (FFBS) [fs]_.

    FFBS draws samples according to

    .. math::

        S_T &\sim p(S_T | y_{1:T}) \\
        S_t \mid S_{t+1} &\sim p(S_{t+1} | S_t) p(S_{t+1} \mid y_{1:T})

    for discrete states in the sequence :math:`S_t`, :math:`t \in \{0, \dots, T\}`
    and observations :math:`y_t`.

    The argument `gamma_0` corresponds to :math:`S_0`, `Gammas` to
    :math:`p(S_{t+1} | S_t)`, and `log_lik` to :math:`\log p(y_t \mid S_t)`.
    The latter is used to derive the forward quantities
    :math:`p(S_{t+1} \mid y_{1:T})`.

    This implementations is most similar to the one in [nr]_, which forgoes
    expensive log-space calculations by using fixed precision bounds and
    re-normalization/scaling.

    Parameters
    ----------
    gamma_0
        The initial state probabilities as an array of base shape ``(M,)``.
    Gamma
        The transition probability matrices.  This array should take the base
        shape ``(N, M, M)``, where ``N`` is the state sequence length and ``M``
        is the number of distinct states.  If ``N`` is ``1``, the single
        transition matrix will broadcast across all elements of the state sequence.
    log_lik
        An array of base shape ``(M, N)`` consisting of the log-likelihood
        values for each state value at each point in the sequence.
    lower_prec_bound
        A number indicating when forward probabilities are too small and need
        to be re-normalized/scaled.

    Returns
    -------
    A tuple containing the tensor representing the FFBS sampled states and the
    `Scan`-generated updates.

    References
    ----------

    .. [fs] Sylvia Fruehwirth-Schnatter, "Markov chain Monte Carlo estimation of classical and dynamic switching and mixture models", Journal of the American Statistical Association 96 (2001): 194--209
    .. [nr] Press, William H., Saul A. Teukolsky, William T. Vetterling, and Brian P. Flannery. 2007. "Numerical Recipes 3rd Edition: The Art of Scientific Computing." Cambridge university press.

    """

    gamma_0 = at.as_tensor(gamma_0)
    log_lik = at.as_tensor(log_lik)

    # Number of observations
    N = log_lik.shape[-1]

    # Number of states
    # M = gamma_0.shape[-1]

    # Initial state probabilities
    gamma_0_normed = gamma_0
    gamma_0_normed /= at.sum(gamma_0)

    # Make sure we have a transition matrix for each element in a state
    # sequence
    Gammas = at.broadcast_to(Gammas, (N,) + Gammas.shape[-2:])

    def forward_step(log_lik_T_n, Gamma_n, alpha_nm1):
        log_lik_n = log_lik_T_n.T
        lik_n = at.exp(log_lik_n - log_lik_n.max())
        alpha_n = at.dot(alpha_nm1, Gamma_n) * lik_n
        alpha_n_sum = at.sum(alpha_n, axis=-1)
        rescaled_alpha_n = at.switch(
            at.lt(alpha_n_sum, lower_prec_bound), alpha_n / lower_prec_bound, alpha_n
        )
        return rescaled_alpha_n

    alphas, _ = aesara.scan(
        forward_step,
        sequences=[log_lik.T, Gammas],
        outputs_info=[{"initial": gamma_0_normed, "taps": [-1]}],
        non_sequences=[],
        n_steps=N,
        strict=True,
        name="forward-pass",
    )

    alphas = alphas.T

    def backward_step(alphas_T_n, Gamma_np1, state_n):
        alphas_n = alphas_T_n.T
        beta_n = alphas_n * Gamma_np1[:, state_n]
        beta_n = beta_n / at.sum(beta_n, axis=-1)
        state_np1 = srng.categorical(beta_n)
        return state_np1

    alpha_N = alphas[..., N - 1]
    beta_N = alpha_N / alpha_N.sum(axis=-1)

    state_np1 = srng.categorical(beta_N)

    state_samples_rev, updates = aesara.scan(
        backward_step,
        sequences=[alphas.T[: N - 1], Gammas[1:]],
        outputs_info=[{"initial": state_np1, "taps": [-1]}],
        non_sequences=[],
        go_backwards=True,
        strict=True,
        name="backward-pass",
    )

    state_samples_rev = at.join(
        0, at.atleast_Nd(state_np1, n=state_samples_rev.type.ndim), state_samples_rev
    )

    return state_samples_rev[::-1], updates



from typing import Dict, Tuple

from aesara.graph.basic import Apply, Variable
from aesara.graph.fg import FunctionGraph
from aesara.tensor.random.utils import RandomStream
from aesara.tensor.var import TensorVariable

import aemcmc.nuts as nuts
from aemcmc.rewriting import (
    SamplerTracker,
    construct_ir_fgraph,
    expand_subsumptions,
    sampler_rewrites_db,
)
from aemcmc.types import Sampler


def construct_sampler(
    obs_rvs_to_values: Dict[TensorVariable, TensorVariable], srng: RandomStream
) -> Tuple[Sampler, Dict[TensorVariable, TensorVariable]]:
    r"""Eagerly construct a sampler for a given set of observed variables and their observations.

    Parameters
    ==========
    obs_rvs_to_values
        A ``dict`` of variables that maps stochastic elements
        (e.g. `RandomVariable`\s) to symbolic `Variable`\s representing their
        observed values.

    Returns
    =======
    A ``dict`` that maps each random variable to its sampler step and
    any updates generated by the sampler steps.
    """

    fgraph, obs_rvs_to_values, memo, new_to_old_rvs = construct_ir_fgraph(
        obs_rvs_to_values
    )

    fgraph.attach_feature(SamplerTracker(srng))

    _ = sampler_rewrites_db.query("+basic").rewrite(fgraph)

    random_vars = tuple(rv for rv in fgraph.outputs if rv not in obs_rvs_to_values)

    discovered_samplers = fgraph.sampler_mappings.rvs_to_samplers

    rvs_to_init_vals = {rv: rv.clone() for rv in random_vars}
    posterior_sample_steps = rvs_to_init_vals.copy()
    # Replace occurrences of observed variables with their observed values
    posterior_sample_steps.update(obs_rvs_to_values)

    # TODO FIXME: Get/extract `Scan`-generated updates
    posterior_updates: Dict[Variable, Variable] = {}

    parameters: Dict[Apply, Tuple[TensorVariable]] = {}
    rvs_without_samplers = set()

    for rv in fgraph.outputs:
        if rv in obs_rvs_to_values:
            continue

        rv_steps = discovered_samplers.get(rv)

        if not rv_steps:
            rvs_without_samplers.add(rv)
            continue

        # TODO FIXME: Just choosing one for now, but we should consider them all.
        step_desc, step, updates = rv_steps.pop()

        # Expand subsumed `DimShuffle`d inputs to `Elemwise`s
        if updates:
            update_keys, update_values = zip(*updates.items())
        else:
            update_keys, update_values = tuple(), tuple()

        sfgraph = FunctionGraph(
            outputs=(step,) + tuple(update_keys) + tuple(update_values),
            clone=False,
            copy_inputs=False,
            copy_orphans=False,
        )

        # Update the other sampled random variables in this step's graph
        sfgraph.replace_all(list(posterior_sample_steps.items()), import_missing=True)

        expand_subsumptions.rewrite(sfgraph)

        step = sfgraph.outputs[0]

        # Update the other sampled random variables in this step's graph
        # (step,) = clone_replace([step], replace=posterior_sample_steps)

        posterior_sample_steps[rv] = step

        if updates:
            keys_offset = len(update_keys) + 1
            update_keys = sfgraph.outputs[1:keys_offset]
            update_values = sfgraph.outputs[keys_offset:]
            updates = dict(zip(update_keys, update_values))
            posterior_updates.update(updates)

    # Use the NUTS sampler for the remaining variables
    if rvs_without_samplers:
        to_sample_rvs = {
            rv: posterior_sample_steps[rv] for rv in list(rvs_without_samplers)
        }
        realized_values = {
            rv: vv
            for rv, vv in posterior_sample_steps.items()
            if rv not in to_sample_rvs
        }
        (nuts_sample_steps, updates, nuts_parameters) = nuts.construct_sampler(
            srng, to_sample_rvs, realized_values
        )

        posterior_sample_steps.update(nuts_sample_steps)
        posterior_updates.update(updates)
        parameters.update(nuts_parameters)

    sampling_steps = {
        new_to_old_rvs[rv]: step
        for rv, step in posterior_sample_steps.items()
        if rv not in obs_rvs_to_values
    }

    return Sampler(sampling_steps, posterior_updates, parameters), {
        new_to_old_rvs[rv]: init_var for rv, init_var in rvs_to_init_vals.items()
    }



import aesara.tensor as at
import numpy as np
from aesara.graph.basic import equal_computations
from aesara.tensor.elemwise import DimShuffle, Elemwise
from cons import car, cdr
from etuples import etuple, etuplize
from unification import unify

from aemcmc.rewriting import (
    SubsumingElemwise,
    construct_ir_fgraph,
    local_elemwise_dimshuffle_subsume,
)


def test_SubsumingElemwise_basics():
    a = at.vector("a")
    b = at.scalar("b")

    x = a * b

    assert isinstance(x.owner.op, Elemwise)
    b_ds = x.owner.inputs[1].owner.op
    assert isinstance(b_ds, DimShuffle)

    ee_mul_op = SubsumingElemwise([a, b], [x])

    assert ee_mul_op != ee_mul_op.clone()
    assert str(ee_mul_op) == "SubsumingElemwise{mul}"

    s = unify(at.mul, ee_mul_op)
    assert s is not False

    assert car(ee_mul_op) == car(x.owner.op)
    assert cdr(ee_mul_op) == cdr(x.owner.op)

    s = unify(etuplize(at.mul), etuplize(ee_mul_op))
    assert s is not False

    ee_et = etuplize(ee_mul_op(a, b))
    x_et = etuple(etuplize(at.mul), a, b)

    s = unify(ee_et, x_et)
    assert s is not False

    # TODO: Consider making this possible
    # s = unify(ee_mul(a, b), x)
    # assert s is not False


def test_local_elemwise_dimshuffle_subsume_basic():
    srng = at.random.RandomStream(2398)

    a = at.vector("a")
    b = srng.normal(0, 1, name="b")

    x = a * b

    node = x.owner
    assert isinstance(node.op, Elemwise)
    b_ds = node.inputs[1].owner.op
    assert isinstance(b_ds, DimShuffle)

    (res,) = local_elemwise_dimshuffle_subsume.transform(None, node)
    assert isinstance(res.owner.op, SubsumingElemwise)
    assert equal_computations(
        [res.owner.op.inner_outputs[0]], [x], res.owner.op.inner_inputs[:2], [a, b]
    )
    assert res.owner.inputs == [a, b]


def test_local_elemwise_dimshuffle_subsume_transpose():
    """Make sure that `local_elemwise_dimshuffle_subsume` is applied selectively."""
    srng = at.random.RandomStream(2398)

    a = at.vector("a")
    # This transpose shouldn't be subsumed, but the one applied to `a` by
    # `Elemwise.make_node` should
    b = srng.normal(at.arange(4).reshape((2, 2)), 1, name="b").T

    x = a * b

    node = x.owner
    assert isinstance(node.op, Elemwise)
    b_ds = node.inputs[1].owner.op
    assert isinstance(b_ds, DimShuffle)

    (res,) = local_elemwise_dimshuffle_subsume.transform(None, node)
    assert isinstance(res.owner.op, SubsumingElemwise)
    assert equal_computations(
        [res.owner.op.inner_outputs[0]], [x], res.owner.op.inner_inputs[:2], [a, b]
    )
    assert res.owner.inputs == [a, b]

    a = at.tensor(np.float64, shape=(None, None, None), name="a")
    # Again, the transpose part shouldn't be subsumed, but the added broadcast
    # dimension should
    b = srng.normal(at.arange(4).reshape((2, 2)), 1, name="b")
    b_ds = b.dimshuffle(("x", 1, 0))

    x = a * b_ds

    node = x.owner
    assert isinstance(node.op, Elemwise)
    b_ds = node.inputs[1].owner.op
    assert isinstance(b_ds, DimShuffle)

    (res,) = local_elemwise_dimshuffle_subsume.transform(None, node)
    assert isinstance(res.owner.op, SubsumingElemwise)
    assert res.owner.inputs[0] == a
    # The input corresponding to `b`/`b_ds` should be equivalent to `b.T`
    assert isinstance(res.owner.inputs[1].owner.op, DimShuffle)
    assert equal_computations([b.T], [res.owner.inputs[1]])


def test_SubsumingElemwise_constant_inputs():
    """Make sure constant inputs are handled correctly by `SubsumingElemwise`."""

    srng = at.random.RandomStream(0)

    s = at.lscalar("s")
    # The `1` is the constant input to a `true_divide` `Elemwise` that should be
    # "subsumed"
    Z = srng.exponential(1, size=s, name="Z")
    mu = 1 / Z
    Y = srng.normal(mu, name="Y")
    y = Y.clone()
    y.name = "y"

    res, *_ = construct_ir_fgraph({Y: y}, clone=False)

    normal_node = res.outputs[1].owner
    subelem_node = normal_node.inputs[3].owner
    assert isinstance(subelem_node.op, SubsumingElemwise)
    assert subelem_node.inputs == [Z]



import aesara.tensor as at
from aesara.graph.rewriting.basic import in2out, node_rewriter
from aesara.graph.rewriting.db import LocalGroupDB
from aesara.graph.rewriting.unify import eval_if_etuple
from aesara.tensor.random.basic import BinomialRV, NegBinomialRV, PoissonRV
from etuples import etuple, etuplize
from kanren import eq, lall, run
from unification import var

from aemcmc.rewriting import sampler_finder_db


def gamma_poisson_conjugateo(observed_val, observed_rv_expr, posterior_expr):
    r"""Produce a goal that represents the application of Bayes theorem
    for a beta prior with a binomial observation model.

    .. math::

        \frac{
            Y \sim \operatorname{Poisson}\left(\lambda\right), \quad
            \lambda \sim \operatorname{Gamma}\left(\alpha, \beta\right)
        }{
            \left(\lambda|Y=y\right) \sim \operatorname{Gamma}\left(\alpha+y, \beta+1\right)
        }


    Parameters
    ----------
    observed_val
        The observed value.
    observed_rv_expr
        An expression that represents the observed variable.
    posterior_exp
        An expression that represents the posterior distribution of the latent
        variable.

    """
    # Gamma-poisson observation model
    alpha_lv, beta_lv = var(), var()
    z_rng_lv = var()
    z_size_lv = var()
    z_type_idx_lv = var()
    z_et = etuple(
        etuplize(at.random.gamma), z_rng_lv, z_size_lv, z_type_idx_lv, alpha_lv, beta_lv
    )
    Y_et = etuple(etuplize(at.random.poisson), var(), var(), var(), z_et)

    # Posterior distribution for p
    new_alpha_et = etuple(etuplize(at.add), alpha_lv, observed_val)
    new_beta_et = etuple(etuplize(at.add), beta_lv, 1)
    z_posterior_et = etuple(
        etuplize(at.random.gamma),
        new_alpha_et,
        new_beta_et,
        rng=z_rng_lv,
        size=z_size_lv,
        dtype=z_type_idx_lv,
    )

    return lall(
        eq(observed_rv_expr, Y_et),
        eq(posterior_expr, z_posterior_et),
    )


@node_rewriter([PoissonRV])
def local_gamma_poisson_posterior(fgraph, node):
    sampler_mappings = getattr(fgraph, "sampler_mappings", None)

    rv_var = node.outputs[1]
    key = ("local_gamma_poisson_posterior", rv_var)

    if sampler_mappings is None or key in sampler_mappings.rvs_seen:
        return None  # pragma: no cover

    q = var()

    rv_et = etuplize(rv_var)

    res = run(None, q, gamma_poisson_conjugateo(rv_var, rv_et, q))
    res = next(res, None)

    if res is None:
        return None  # pragma: no cover

    gamma_rv = rv_et[-1].evaled_obj
    gamma_posterior = eval_if_etuple(res)

    sampler_mappings.rvs_to_samplers.setdefault(gamma_rv, []).append(
        ("local_gamma_poisson_posterior", gamma_posterior, None)
    )
    sampler_mappings.rvs_seen.add(key)

    return rv_var.owner.outputs


def beta_binomial_conjugateo(observed_val, observed_rv_expr, posterior_expr):
    r"""Produce a goal that represents the application of Bayes theorem
    for a beta prior with a binomial observation model.

    .. math::

        \frac{
            Y \sim \operatorname{Binom}\left(N, p\right), \quad
            p \sim \operatorname{Beta}\left(\alpha, \beta\right)
        }{
            \left(p|Y=y\right) \sim \operatorname{Beta}\left(\alpha+y, \beta+N-y\right)
        }


    Parameters
    ----------
    observed_val
        The observed value.
    observed_rv_expr
        An expression that represents the observed variable.
    posterior_exp
        An expression that represents the posterior distribution of the latent
        variable.

    """
    # Beta-binomial observation model
    alpha_lv, beta_lv = var(), var()
    p_rng_lv = var()
    p_size_lv = var()
    p_type_idx_lv = var()
    p_et = etuple(
        etuplize(at.random.beta), p_rng_lv, p_size_lv, p_type_idx_lv, alpha_lv, beta_lv
    )
    n_lv = var()
    Y_et = etuple(etuplize(at.random.binomial), var(), var(), var(), n_lv, p_et)

    # Posterior distribution for p
    new_alpha_et = etuple(etuplize(at.add), alpha_lv, observed_val)
    new_beta_et = etuple(
        etuplize(at.sub), etuple(etuplize(at.add), beta_lv, n_lv), observed_val
    )
    p_posterior_et = etuple(
        etuplize(at.random.beta),
        new_alpha_et,
        new_beta_et,
        rng=p_rng_lv,
        size=p_size_lv,
        dtype=p_type_idx_lv,
    )

    return lall(
        eq(observed_rv_expr, Y_et),
        eq(posterior_expr, p_posterior_et),
    )


@node_rewriter([BinomialRV])
def local_beta_binomial_posterior(fgraph, node):
    sampler_mappings = getattr(fgraph, "sampler_mappings", None)

    rv_var = node.outputs[1]
    key = ("local_beta_binomial_posterior", rv_var)

    if sampler_mappings is None or key in sampler_mappings.rvs_seen:
        return None  # pragma: no cover

    q = var()

    rv_et = etuplize(rv_var)

    res = run(None, q, beta_binomial_conjugateo(rv_var, rv_et, q))
    res = next(res, None)

    if res is None:
        return None  # pragma: no cover

    beta_rv = rv_et[-1].evaled_obj
    beta_posterior = eval_if_etuple(res)

    sampler_mappings.rvs_to_samplers.setdefault(beta_rv, []).append(
        ("local_beta_binomial_posterior", beta_posterior, None)
    )
    sampler_mappings.rvs_seen.add(key)

    return rv_var.owner.outputs


def beta_negative_binomial_conjugateo(observed_val, observed_rv_expr, posterior_expr):
    r"""Produce a goal that represents the application of Bayes theorem
    for a beta prior with a negative binomial observation model.

    .. math::

        \frac{
            Y \sim \operatorname{NB}\left(k, p\right), \quad
            p \sim \operatorname{Beta}\left(\alpha, \beta\right)
        }{
            \left(p \mid Y=y\right) \sim \operatorname{Beta}\left(\alpha + \sum^{n}_{i=1} y_i, \beta + k n\right)
        }
        where :math:`k` is the number of successes before experiment ended.


    Parameters
    ----------
    observed_val
        The observed value.
    observed_rv_expr
        An expression that represents the observed variable.
    posterior_exp
        An expression that represents the posterior distribution of the latent
        variable.

    """
    # beta-negative_binomial observation model
    alpha_lv, beta_lv = var(), var()
    p_rng_lv = var()
    p_size_lv = var()
    p_type_idx_lv = var()
    p_et = etuple(
        etuplize(at.random.beta), p_rng_lv, p_size_lv, p_type_idx_lv, alpha_lv, beta_lv
    )
    n_lv = var()  # success
    Y_et = etuple(
        etuplize(at.random.negative_binomial), var(), var(), var(), n_lv, p_et
    )

    new_alpha_et = etuple(etuplize(at.add), alpha_lv, observed_val)
    new_beta_et = etuple(etuplize(at.add), beta_lv, n_lv)
    p_posterior_et = etuple(
        etuplize(at.random.beta),
        new_alpha_et,
        new_beta_et,
        rng=p_rng_lv,
        size=p_size_lv,
        dtype=p_type_idx_lv,
    )

    return lall(
        eq(observed_rv_expr, Y_et),
        eq(posterior_expr, p_posterior_et),
    )


@node_rewriter([NegBinomialRV])
def local_beta_negative_binomial_posterior(fgraph, node):
    sampler_mappings = getattr(fgraph, "sampler_mappings", None)

    rv_var = node.outputs[1]
    key = ("local_beta_negative_binomial_posterior", rv_var)

    if sampler_mappings is None or key in sampler_mappings.rvs_seen:
        return None  # pragma: no cover

    q = var()

    rv_et = etuplize(rv_var)

    res = run(None, q, beta_negative_binomial_conjugateo(rv_var, rv_et, q))
    res = next(res, None)

    if res is None:
        return None  # pragma: no cover

    beta_rv = rv_et[-1].evaled_obj
    beta_posterior = eval_if_etuple(res)

    sampler_mappings.rvs_to_samplers.setdefault(beta_rv, []).append(
        ("local_beta_negative_binomial_posterior", beta_posterior, None)
    )
    sampler_mappings.rvs_seen.add(key)

    return rv_var.owner.outputs


conjugates_db = LocalGroupDB(apply_all_rewrites=True)
conjugates_db.name = "conjugates_db"
conjugates_db.register("beta_binomial", local_beta_binomial_posterior, "basic")
conjugates_db.register("gamma_poisson", local_gamma_poisson_posterior, "basic")
conjugates_db.register(
    "negative_binomial", local_beta_negative_binomial_posterior, "basic"
)


sampler_finder_db.register(
    "conjugates", in2out(conjugates_db.query("+basic"), name="gibbs"), "basic"
)



import aesara
import aesara.tensor as at
import numpy as np
from aesara.sparse.basic import _is_sparse_variable, dense_from_sparse
from aesara.tensor.random import RandomStream
from aesara.tensor.random.op import RandomVariable
from aesara.tensor.var import TensorVariable
from polyagamma import random_polyagamma


class PolyaGammaRV(RandomVariable):
    name = "polyagamma"
    ndim_supp = 0
    ndims_params = [0, 0]
    dtype = "floatX"
    _print_name = ("PG", "\\operatorname{PG}")

    def __call__(self, h=1.0, z=0.0, size=None, **kwargs):
        return super().__call__(h, z, size=size, **kwargs)

    @classmethod
    def rng_fn(cls, rng, h, z, size=None):
        bg = rng._bit_generator if isinstance(rng, np.random.RandomState) else rng
        return random_polyagamma(h, z, size=size, random_state=bg).astype(
            aesara.config.floatX
        )


polyagamma = PolyaGammaRV()


def multivariate_normal_rue2005(rng, b, Q):
    r"""Sample from a multivariate normal distribution.

    More specifically, this function draws a sample from the following distribution:

    .. math::

        \operatorname{N}\left( Q^{-1} b, Q^{-1} \right)

    It uses the algorithm described in [1]_, which is suitable for cases in
    which the number of regression coefficients is significantly less than the
    number of data points.

    References
    ----------
    .. [1] Rue, H. and Held, L. (2005), Gaussian Markov Random Fields, Boca
          Raton: Chapman & Hall/CRC.
    """
    if _is_sparse_variable(Q):
        Q = dense_from_sparse(Q)
    L = at.linalg.cholesky(Q)
    w = at.slinalg.solve_triangular(L, b, lower=True)
    u = at.slinalg.solve_triangular(L.T, w, lower=False)
    z = rng.standard_normal(size=L.shape[0])
    z.owner.outputs[0].name = "z_rng"
    v = at.slinalg.solve_triangular(L.T, z, lower=False)
    return u + v


def multivariate_normal_cong2017(
    rng: RandomStream,
    A: TensorVariable,
    omega: TensorVariable,
    phi: TensorVariable,
    t: TensorVariable,
) -> TensorVariable:
    r"""Sample from a multivariate normal distribution with a structured mean and covariance.

    As described in Example 4 [page 17] of [1]_, The covariance of this normal
    distribution should be decomposable into a sum of a positive-definite matrix
    and a low-rank symmetric matrix such that:

    .. math::

         \operatorname{N}\left(\Lambda^{-1} \Phi^{\top} \Omega t, \Lambda^{-1}\right)

    where

    .. math::

        \Lambda = A + \Phi^{\top} \Omega \Phi

    and :math:`A` is the positive-definite part and
    :math:`\Phi^{\top} \Omega \Phi` is the eigen-factorization of
    the low-rank part of the "structured" covariance.

    Parameters
    ----------
    rng
        The random number generating object to be used during sampling.
    A
        The entries of the diagonal elements of the positive-definite part of
        the structured covariance.
    omega
        The elements of the diagonal matrix in the eigen-decomposition of the
        low-rank part of the structured covariance of the multivariate normal
        distribution.
    phi
        A matrix containing the eigenvectors of the eigen-decomposition of the
        low-rank part of the structured covariance of the normal distribution.
    t
        A 1D array whose length is the number of eigenvalues of the low-rank
        part of the structured covariance.

    Notes
    -----
    This algorithm is suitable for high-dimensional regression problems and the
    runtime scales linearly with the number of regression coefficients. This
    implementation assumes that `A` and `omega` are diagonal matrices and
    the parameters `A` and `omega` are expected to be vectors that contain
    diagonal entries of the respective matrices they represent.

    Note the the algorithm described in [2]_ is a special case when `omega` is
    the identity matrix. Samples from the algorithm described in [2]_ can be
    generated using the following:

    .. code::

        rng = at.random.RandomStream(12345)
        X = rng.standard_normal(size=(50, 100))
        phi = at.linalg.cholesky(X @ X.T).T
        D = rng.random(phi.shape[1])
        alpha = rng.random(phi.shape[0])
        multivariate_normal_cong2017(rng, 1 / D, at.ones(phi.shape[0]), phi, alpha)

    References
    ----------
    .. [1] Cong, Yulai; Chen, Bo; Zhou, Mingyuan. Fast Simulation of Hyperplane-
          Truncated Multivariate Normal Distributions. Bayesian Anal. 12 (2017),
          no. 4, 1017--1037. doi:10.1214/17-BA1052.
    .. [2] Bhattacharya, A., Chakraborty, A., and Mallick, B. K. (2016).
          “Fast sampling with Gaussian scale mixture priors in high-dimensional
          regression.” Biometrika, 103(4): 985.033
    """
    A_inv = 1 / A
    a_rows = A.shape[0]
    z = rng.standard_normal(size=a_rows + omega.shape[0])
    z.owner.outputs[0].name = "z_rng"
    y1 = at.sqrt(A_inv) * z[:a_rows]
    y2 = (1 / at.sqrt(omega)) * z[a_rows:]
    Ainv_phi = A_inv[:, None] * phi.T
    # NOTE: B is equivalent to B = phi @ Ainv_phi + at.diag(1 / omega), but
    # may be sligthly more efficient since we do not allocate memory for the
    # full inverse of omega but instead use just the diagonal elements.
    B = phi @ Ainv_phi
    indices = at.arange(B.shape[0])
    B = at.subtensor.set_subtensor(
        B[indices, indices],
        at.diag(B) + 1.0 / omega,
    )
    alpha = at.linalg.solve(B, t - phi @ y1 - y2, assume_a="pos")
    return y1 + Ainv_phi @ alpha



import aesara
import aesara.tensor as at
import aesara.tensor.random as atr
import numpy as np
import pytest
import scipy.stats as sp

from aemcmc.ffbs import ffbs_step


@pytest.mark.parametrize("first_is_nonzero", [True, False])
def test_ffbs_one_state_likelihood(first_is_nonzero):
    """Test for categorical emissions that assign non-zero likelihood to only one state."""

    Gammas = np.array([[[0.9, 0.1], [0.1, 0.9]]])
    gamma_0 = np.r_[0.5, 0.5]

    N = 10

    if first_is_nonzero:
        log_lik = at.stack([at.broadcast_to(0.0, (N,)), at.broadcast_to(-np.inf, (N,))])
    else:
        log_lik = at.stack([at.broadcast_to(-np.inf, (N,)), at.broadcast_to(0.0, (N,))])

    srng = atr.RandomStream(seed=2032)
    step_at, updates = ffbs_step(gamma_0, Gammas, log_lik, srng)
    step_fn = aesara.function([], step_at, updates=updates)
    res = step_fn()

    if first_is_nonzero:
        assert np.all(res == 0)
    else:
        assert np.all(res == 1)


def test_ffbs_poisson_states():
    """Test well-separated Poisson emissions."""
    N = 10
    Gammas = np.array([[[0.9, 0.1], [0.1, 0.9]]])
    gamma_0 = np.r_[0.5, 0.5]

    rng = np.random.default_rng(2032)
    seq = rng.choice(2, size=N)
    obs = np.where(
        np.logical_not(seq),
        np.random.poisson(1, N),
        np.random.poisson(50, N),
    )
    log_lik = np.stack(
        [sp.poisson.logpmf(obs, 1), sp.poisson.logpmf(obs, 50)],
    )

    # TODO FIXME: This is a fairly arbitrary check
    assert np.mean(np.abs(log_lik.argmax(0) - seq)) < 1e-2

    srng = atr.RandomStream(seed=2032)
    step_at, updates = ffbs_step(gamma_0, Gammas, log_lik, srng)
    step_fn = aesara.function([], step_at, updates=updates)
    res = step_fn()

    # TODO FIXME: This is a fairly arbitrary check
    assert np.mean(np.abs(res - seq)) < 1e-2


def test_ffbs_strict_alternating_transitions():
    """Test time-varying transition matrices that specify strictly alternating states--except for the second-to-last one."""
    Gammas = np.stack(
        [
            np.array([[0.0, 1.0], [1.0, 0.0]]),
            np.array([[0.0, 1.0], [1.0, 0.0]]),
            np.array([[1.0, 0.0], [0.0, 1.0]]),
            np.array([[0.0, 1.0], [1.0, 0.0]]),
        ],
        axis=0,
    )

    gamma_0 = np.r_[1.0, 0.0]

    log_lik = np.tile(np.r_[np.log(0.9), np.log(0.1)], (4, 1))
    log_lik[::2] = log_lik[::2][:, ::-1]
    log_lik = log_lik.T

    srng = atr.RandomStream(seed=2032)
    step_at, updates = ffbs_step(gamma_0, Gammas, log_lik, srng)
    step_fn = aesara.function([], step_at, updates=updates)
    res = step_fn()

    assert np.array_equal(res, np.r_[1, 0, 0, 1])



import aesara.tensor as at
import pytest

from aemcmc.utils import ModelInfo


def test_ModelInfo_maps():
    """Test maps generated by `ModelInfo`'s constructor."""
    srng = at.random.RandomStream()

    beta_rv = srng.normal(0, 1, name="beta")
    beta_vv = beta_rv.clone()
    beta_vv.name = "beta_vv"

    X = at.matrix("X")

    mu = at.exp(X.dot(beta_rv))
    mu.name = "mu"

    Y_rv = srng.poisson(mu, name="Y")
    y_vv = Y_rv.clone()
    y_vv.name = "y"

    observed_rvs = (Y_rv,)
    rvs_to_values = {Y_rv: y_vv, beta_rv: beta_vv}
    deterministic_vars = (mu,)

    amodel = ModelInfo(observed_rvs, rvs_to_values, deterministic_vars, {})

    assert amodel.values_to_rvs == {y_vv: Y_rv, beta_vv: beta_rv}
    assert amodel.names_to_vars == {
        "Y": Y_rv,
        "beta": beta_rv,
        "beta_vv": beta_vv,
        "mu": mu,
        "y": y_vv,
    }
    assert amodel.observed_values == (y_vv,)
    assert amodel.unobserved_rvs == (beta_rv,)
    assert amodel.unobserved_values == (beta_vv,)
    assert amodel.updates == {}
    assert set(amodel.names_to_vars.values()) == set((Y_rv, beta_rv, mu, y_vv, beta_vv))

    smodel = ModelInfo(observed_rvs, rvs_to_values)
    assert smodel.deterministic_vars == ()
    assert smodel.updates == {}
    assert set(smodel.names_to_vars.values()) == set((Y_rv, beta_rv, y_vv, beta_vv))


def test_ModelInfo_errors():
    with pytest.raises(ValueError, match=".*non-empty names.*"):
        ModelInfo((), {at.vector("a"): at.vector()}, (), {})

    with pytest.raises(ValueError, match=".*unique names.*"):
        ModelInfo((), {at.vector("a"): at.vector("a")}, (), {})



import aesara
import aesara.tensor as at
import numpy as np
import pytest
import scipy.special
from aesara.graph.basic import equal_computations
from aesara.graph.rewriting.utils import rewrite_graph
from aesara.tensor.random.utils import RandomStream
from scipy.linalg import toeplitz

from aemcmc.gibbs import (
    bern_normal_posterior,
    bern_sigmoid_dot_match,
    gamma_match,
    horseshoe_match,
    horseshoe_posterior,
    nbinom_dispersion_posterior,
    nbinom_normal_posterior,
    normal_regression_posterior,
    sample_CRT,
)
from aemcmc.rewriting import SamplerTracker, construct_ir_fgraph, sampler_rewrites_db


@pytest.fixture
def srng():
    return RandomStream(1234)


def test_horseshoe_match(srng):
    size = at.lscalar("size")
    # Vector tau
    tau_rv = srng.halfcauchy(0, 1, size=1, name="tau")
    lmbda_rv = srng.halfcauchy(0, 1, size=size, name="lambda")
    beta_rv = srng.normal(0, lmbda_rv * tau_rv, size=size, name="beta")

    fgraph, _, memo, _ = construct_ir_fgraph({beta_rv: beta_rv})
    beta_rv = fgraph.outputs[-1]

    lambda_res, tau_res = horseshoe_match(beta_rv)

    assert lambda_res is memo[lmbda_rv]
    assert tau_res is memo[tau_rv]

    # Scalar tau
    tau_rv = srng.halfcauchy(0, 1, name="tau")
    lmbda_rv = srng.halfcauchy(0, 1, size=size, name="lambda")
    beta_rv = srng.normal(0, lmbda_rv * tau_rv, size=size, name="beta")

    fgraph, _, memo, _ = construct_ir_fgraph({beta_rv: beta_rv})
    beta_rv = fgraph.outputs[-1]

    lambda_res, tau_res = horseshoe_match(beta_rv)

    assert lambda_res is memo[lmbda_rv]
    # `tau_res` should've had its `DimShuffle` lifted, so it's not identical to `tau_rv`
    assert isinstance(tau_res.owner.op, type(tau_rv.owner.op))
    assert tau_res.type.ndim == 0

    beta_rv = srng.normal(0, 1)
    with pytest.raises(ValueError):
        horseshoe_match(beta_rv)

    size = at.scalar("size", dtype="int32")
    tau_rv = srng.halfcauchy(0, 1, size=1)
    lmbda_rv = srng.normal(0, 1, size=size)
    beta_rv = srng.normal(0, tau_rv * lmbda_rv, size=size)
    with pytest.raises(ValueError):
        horseshoe_match(beta_rv)

    size = at.scalar("size", dtype="int32")
    tau_rv = srng.normal(0, 1, size=1)
    lmbda_rv = srng.halfcauchy(0, 1, size=size)
    beta_rv = srng.normal(0, tau_rv * lmbda_rv, size=size)
    with pytest.raises(ValueError):
        horseshoe_match(beta_rv)

    size = at.scalar("size", dtype="int32")
    tau_rv = srng.halfcauchy(0, 1, size=size)
    lmbda_rv = srng.halfcauchy(0, 1, size=size)

    beta_rv = srng.normal(0, tau_rv * lmbda_rv, size=size)
    with pytest.raises(ValueError):
        horseshoe_match(beta_rv)


@pytest.mark.parametrize(
    "N, p, rtol",
    [
        (50, 10, 0.5),
        (50, 75, 0.5),
    ],
)
def test_normal_horseshoe_sampler(srng, N, p, rtol):
    """Check the results of a normal regression model with a Horseshoe prior.

    This test example is modified from section 3.2 of Makalic & Schmidt (2016)

    """
    rng = np.random.default_rng(9420)

    true_beta = np.array([5, 3, 3, 1, 1] + [0] * (p - 5))
    S = toeplitz(0.5 ** np.arange(p))
    X = rng.multivariate_normal(np.zeros(p), cov=S, size=N)
    y = rng.normal(X @ true_beta, np.ones(N))

    tau_rv = srng.halfcauchy(0, 1)
    lambda_rv = srng.halfcauchy(0, 1, size=p)

    tau_vv = tau_rv.clone()
    tau_vv.name = "tau"
    lambda_vv = lambda_rv.clone()
    lambda_vv.name = "lambda"

    beta_post = normal_regression_posterior(
        srng, np.ones(N), at.reciprocal(tau_vv * lambda_vv), at.as_tensor(X), y
    )

    lambda_post, tau_post = horseshoe_posterior(srng, beta_post, 1.0, lambda_vv, tau_vv)

    outputs = (beta_post, lambda_post, tau_post)
    sample_fn = aesara.function((tau_vv, lambda_vv), outputs)

    beta_post_vals = []
    lambda_post_val, tau_post_val = np.ones(p), 1.0
    for i in range(3000):
        beta_post_val, lambda_post_val, tau_post_val = sample_fn(
            tau_post_val, lambda_post_val
        )
        beta_post_vals += [beta_post_val]
        assert np.all(tau_post_val >= 0)
        assert np.all(lambda_post_val >= 0)

    beta_post_median = np.median(beta_post_vals[100::2], axis=0)
    assert np.allclose(beta_post_median[:5], true_beta[:5], atol=1e-1, rtol=rtol)


@pytest.mark.parametrize(
    "h_val, y_val",
    [
        (1.0, 0),
        (1.0, 1),
        (1.0, 20),
        (10.0, 20),
    ],
)
def test_sample_CRT_mean(srng, h_val, y_val):
    y = at.lvector("y")
    h = at.scalar("h")

    crt_res, updates = sample_CRT(srng, y, h)

    crt_fn = aesara.function([y, h], crt_res, updates=updates)

    y_vals = np.repeat(y_val, 5000).astype(np.int64)

    crt_vals = crt_fn(y_vals, h_val)

    crt_mean_val = crt_vals.mean()
    crt_exp_val = h_val * (
        scipy.special.digamma(h_val + y_val) - scipy.special.digamma(h_val)
    )

    assert np.allclose(crt_mean_val, crt_exp_val, rtol=1e-1)


def test_nbinom_normal_posterior(srng):
    M = 10
    N = 50

    true_h = 100
    true_beta = np.array([2, 0.02, 0.2, 0.1, 1] + [0.0] * (M - 5))
    S = toeplitz(0.5 ** np.arange(M))
    X_at = srng.multivariate_normal(np.zeros(M), cov=S, size=N)
    p_at = at.sigmoid(-(X_at.dot(true_beta)))
    X, p, y = aesara.function([], [X_at, p_at, srng.nbinom(true_h, p_at)])()

    beta_vv = at.vector("beta")
    beta_post = nbinom_normal_posterior(
        srng, beta_vv, 200 * np.ones(M), at.as_tensor(X), true_h, y
    )

    beta_post_fn = aesara.function([beta_vv], beta_post)

    beta_post_vals = []
    beta_post_val = np.zeros(M)
    for i in range(1000):
        beta_post_val = beta_post_fn(beta_post_val)
        beta_post_vals += [beta_post_val]

    beta_post_mean = np.mean(beta_post_vals, axis=0)
    assert np.allclose(beta_post_mean, true_beta, atol=1e-1)


def test_nbinom_dispersion_posterior(srng):
    M = 10
    N = 100

    true_h = 100
    true_beta = np.array([2, 0.02, 0.2, 0.1, 1] + [0.1] * (M - 5))
    S = toeplitz(0.5 ** np.arange(M))
    X = srng.multivariate_normal(np.zeros(M), cov=S, size=N)
    p_at = at.sigmoid(-(X.dot(true_beta)))
    p, y = aesara.function([], [p_at, srng.nbinom(true_h, p_at)])()

    a_val = 70.0
    b_val = 0.9
    a = at.as_tensor(a_val)
    b = at.as_tensor(b_val)

    h_samples, h_updates = aesara.scan(
        lambda last_h: nbinom_dispersion_posterior(srng, last_h, p, a, b, y),
        outputs_info=[at.as_tensor(90.0, dtype=np.float64)],
        n_steps=1000,
    )

    h_mean_fn = aesara.function([], h_samples.mean(), updates=h_updates)

    h_mean_val = h_mean_fn()

    # Make sure that the average posterior `h` values have increased relative
    # to the prior mean
    assert h_mean_val > a_val / b_val

    # Make sure the posterior values are near the "true" value
    assert np.allclose(h_mean_val, true_h, rtol=1e-1)


def test_bern_sigmoid_dot_match(srng):
    X = at.matrix("X")

    beta_rv = srng.normal(0, 1, size=X.shape[1], name="beta")
    eta = X @ beta_rv
    p = at.sigmoid(-eta)
    Y_rv = srng.bernoulli(p)

    Y_rv = rewrite_graph(Y_rv)

    assert bern_sigmoid_dot_match(Y_rv)

    beta = at.vector("beta")
    X = at.matrix("X")
    Y = X @ beta

    with pytest.raises(ValueError):
        bern_sigmoid_dot_match(Y)

    X = at.matrix("X")
    beta_rv = srng.normal(0, 1, name="beta")
    eta = X @ beta_rv
    p = at.sigmoid(2 * eta)
    Y_rv = srng.bernoulli(p)

    with pytest.raises(ValueError):
        bern_sigmoid_dot_match(Y_rv)


def test_bern_normal_posterior(srng):
    M = 10
    N = 100

    true_beta = np.array([3, 2, 1, 0.5, 0.05] + [0.0] * (M - 5))
    S = toeplitz(0.5 ** np.arange(M))
    X_at = srng.multivariate_normal(np.zeros(M), cov=S, size=N)
    p_at = at.sigmoid(X_at.dot(true_beta))
    X, p, y = aesara.function([], [X_at, p_at, srng.bernoulli(p_at)])()

    beta_vv = at.vector("beta")
    beta_post = bern_normal_posterior(srng, beta_vv, np.ones(M), at.as_tensor(X), y)

    beta_post_fn = aesara.function([beta_vv], beta_post)

    beta_post_vals = []
    beta_post_val = np.zeros(M)
    for i in range(1000):
        beta_post_val = beta_post_fn(beta_post_val)
        beta_post_vals += [beta_post_val]

    beta_post_mean = np.mean(beta_post_vals, axis=0)
    assert np.allclose(beta_post_mean, true_beta, atol=0.5)


def test_gamma_match(srng):
    beta_rv = srng.normal(0, 1)

    with pytest.raises(ValueError):
        gamma_match(beta_rv)

    a = at.scalar("a")
    b = at.scalar("b")
    beta_rv = srng.gamma(a, b)

    beta_rv = rewrite_graph(beta_rv)

    a_m, b_m = gamma_match(beta_rv)

    assert a_m is a

    b_exp = at.as_tensor(1.0, dtype="floatX") / b
    assert equal_computations([b_m], [b_exp])


def test_nbinom_logistic_horseshoe_finders():
    """Make sure `nbinom_logistic_finder` and `normal_horseshoe_finder` work."""
    srng = RandomStream(0)

    X = at.matrix("X")

    # Horseshoe `beta_rv`
    tau_rv = srng.halfcauchy(0, 1, name="tau")
    lmbda_rv = srng.halfcauchy(0, 1, size=X.shape[1], name="lambda")
    beta_rv = srng.normal(0, lmbda_rv * tau_rv, size=X.shape[1], name="beta")

    a = at.scalar("a")
    b = at.scalar("b")
    h_rv = srng.gamma(a, b, name="h")

    # Negative-binomial regression
    eta = X @ beta_rv
    p = at.sigmoid(-eta)
    Y_rv = srng.nbinom(h_rv, p, name="Y")

    y_vv = Y_rv.clone()
    y_vv.name = "y"

    fgraph, obs_rvs_to_values, memo, new_to_old_rvs = construct_ir_fgraph({Y_rv: y_vv})

    fgraph.attach_feature(SamplerTracker(srng))

    _ = sampler_rewrites_db.query("+basic").rewrite(fgraph)

    discovered_samplers = fgraph.sampler_mappings.rvs_to_samplers
    discovered_samplers = {
        new_to_old_rvs[rv]: discovered_samplers.get(rv)
        for rv in fgraph.outputs
        if rv not in obs_rvs_to_values
    }

    assert len(discovered_samplers) == 4

    assert discovered_samplers[tau_rv][0][0] == "normal_horseshoe_finder"
    assert discovered_samplers[lmbda_rv][0][0] == "normal_horseshoe_finder"
    assert discovered_samplers[beta_rv][0][0] == "nbinom_logistic_finder"
    assert discovered_samplers[h_rv][0][0] == "nbinom_logistic_finder"


def test_bern_logistic_horseshoe_finders():
    """Make sure `bern_logistic_finder` and `normal_horseshoe_finder` work."""
    srng = RandomStream(0)

    X = at.matrix("X")

    # Horseshoe `beta_rv`
    tau_rv = srng.halfcauchy(0, 1, name="tau")
    lmbda_rv = srng.halfcauchy(0, 1, size=X.shape[1], name="lambda")
    beta_rv = srng.normal(0, lmbda_rv * tau_rv, size=X.shape[1], name="beta")

    # Negative-binomial regression
    eta = X @ beta_rv
    p = at.sigmoid(-eta)
    Y_rv = srng.bernoulli(p, name="Y")

    y_vv = Y_rv.clone()
    y_vv.name = "y"

    fgraph, obs_rvs_to_values, memo, new_to_old_rvs = construct_ir_fgraph({Y_rv: y_vv})

    fgraph.attach_feature(SamplerTracker(srng))

    _ = sampler_rewrites_db.query("+basic").rewrite(fgraph)

    discovered_samplers = fgraph.sampler_mappings.rvs_to_samplers
    discovered_samplers = {
        new_to_old_rvs[rv]: discovered_samplers.get(rv)
        for rv in fgraph.outputs
        if rv not in obs_rvs_to_values
    }

    assert len(discovered_samplers) == 3

    assert discovered_samplers[tau_rv][0][0] == "normal_horseshoe_finder"
    assert discovered_samplers[lmbda_rv][0][0] == "normal_horseshoe_finder"
    assert discovered_samplers[beta_rv][0][0] == "bern_logistic_finder"



import aesara
import aesara.tensor as at
import pytest
from aesara.graph.rewriting.unify import eval_if_etuple
from aesara.tensor.random import RandomStream
from kanren import run
from unification import var

from aemcmc.conjugates import (
    beta_binomial_conjugateo,
    beta_negative_binomial_conjugateo,
    gamma_poisson_conjugateo,
)


def test_gamma_poisson_conjugate_contract():
    """Produce the closed-form posterior for the poisson observation model with
    a gamma prior.

    """
    srng = RandomStream(0)

    alpha_tt = at.scalar("alpha")
    beta_tt = at.scalar("beta")
    z_rv = srng.gamma(alpha_tt, beta_tt)

    Y_rv = srng.poisson(z_rv)
    y_vv = Y_rv.clone()
    y_vv.name = "y"

    q_lv = var()
    (posterior_expr,) = run(1, q_lv, gamma_poisson_conjugateo(y_vv, Y_rv, q_lv))
    posterior = eval_if_etuple(posterior_expr)

    assert isinstance(posterior.owner.op, type(at.random.gamma))


@pytest.mark.xfail(
    reason="Op.__call__ does not dispatch to Op.make_node for some RandomVariable and etuple evaluation returns an error"
)
def test_gamma_poisson_conjugate_expand():
    """Expand a contracted beta-binomial observation model."""

    srng = RandomStream(0)

    alpha_tt = at.scalar("alpha")
    beta_tt = at.scalar("beta")
    y_vv = at.iscalar("y")
    Y_rv = srng.gamma(alpha_tt + y_vv, beta_tt + 1)

    e_lv = var()
    (expanded_expr,) = run(1, e_lv, gamma_poisson_conjugateo(e_lv, y_vv, Y_rv))
    expanded = eval_if_etuple(expanded_expr)

    assert isinstance(expanded.owner.op, type(at.random.gamma))


def test_beta_binomial_conjugate_contract():
    """Produce the closed-form posterior for the binomial observation model with
    a beta prior.

    """
    srng = RandomStream(0)

    alpha_tt = at.scalar("alpha")
    beta_tt = at.scalar("beta")
    p_rv = srng.beta(alpha_tt, beta_tt, name="p")

    n_tt = at.iscalar("n")
    Y_rv = srng.binomial(n_tt, p_rv)
    y_vv = Y_rv.clone()
    y_vv.tag.name = "y"

    q_lv = var()
    (posterior_expr,) = run(1, q_lv, beta_binomial_conjugateo(y_vv, Y_rv, q_lv))
    posterior = eval_if_etuple(posterior_expr)

    assert isinstance(posterior.owner.op, type(at.random.beta))

    # Build the sampling function and check the results on limiting cases.
    sample_fn = aesara.function((alpha_tt, beta_tt, y_vv, n_tt), posterior)
    assert sample_fn(1.0, 1.0, 1000, 1000) == pytest.approx(
        1.0, abs=0.01
    )  # only successes
    assert sample_fn(1.0, 1.0, 0, 1000) == pytest.approx(0.0, abs=0.01)  # zero success


@pytest.mark.xfail(
    reason="Op.__call__ does not dispatch to Op.make_node for some RandomVariable and etuple evaluation returns an error"
)
def test_beta_binomial_conjugate_expand():
    """Expand a contracted beta-binomial observation model."""

    srng = RandomStream(0)

    alpha_tt = at.scalar("alpha")
    beta_tt = at.scalar("beta")
    y_vv = at.iscalar("y")
    n_tt = at.iscalar("n")
    Y_rv = srng.beta(alpha_tt + y_vv, beta_tt + n_tt - y_vv)

    e_lv = var()
    (expanded_expr,) = run(1, e_lv, beta_binomial_conjugateo(e_lv, y_vv, Y_rv))
    expanded = eval_if_etuple(expanded_expr)

    assert isinstance(expanded.owner.op, type(at.random.beta))


def test_beta_negative_binomial_conjugate_contract():
    """Produce the closed-form posterior for the binomial observation model with
    a beta prior.

    """
    srng = RandomStream(0)

    alpha_tt = at.scalar("alpha")
    beta_tt = at.scalar("beta")
    p_rv = srng.beta(alpha_tt, beta_tt, name="p")

    n_tt = at.iscalar("n")
    Y_rv = srng.negative_binomial(n_tt, p_rv)
    y_vv = Y_rv.clone()
    y_vv.tag.name = "y"

    q_lv = var()
    (posterior_expr,) = run(
        1, q_lv, beta_negative_binomial_conjugateo(y_vv, Y_rv, q_lv)
    )
    posterior = eval_if_etuple(posterior_expr)

    assert isinstance(posterior.owner.op, type(at.random.beta))

    # Build the sampling function and check the results on limiting cases.
    sample_fn = aesara.function((alpha_tt, beta_tt, y_vv, n_tt), posterior)
    assert sample_fn(1.0, 1.0, 1000, 0) == pytest.approx(
        1.0, abs=0.01
    )  # only successes
    assert sample_fn(1.0, 1.0, 0, 1000) == pytest.approx(0.0, abs=0.01)  # no success


@pytest.mark.xfail(
    reason="Op.__call__ does not dispatch to Op.make_node for some RandomVariable and etuple evaluation returns an error"
)
def test_beta_negative_binomial_conjugate_expand():
    """Expand a contracted beta-binomial observation model."""

    srng = RandomStream(0)

    alpha_tt = at.scalar("alpha")
    beta_tt = at.scalar("beta")
    y_vv = at.iscalar("y")
    n_tt = at.iscalar("n")
    Y_rv = srng.beta(alpha_tt + y_vv, beta_tt + n_tt)

    e_lv = var()
    (expanded_expr,) = run(1, e_lv, beta_negative_binomial_conjugateo(e_lv, y_vv, Y_rv))
    expanded = eval_if_etuple(expanded_expr)

    assert isinstance(expanded.owner.op, type(at.random.beta))



from collections import defaultdict
from dataclasses import dataclass, field
from typing import Dict, List, Optional, Tuple

from aesara.compile.builders import OpFromGraph
from aesara.graph.basic import Variable
from aesara.tensor.var import TensorVariable


class SamplingStep(OpFromGraph):
    """An AeMCMC sampling step.

    Sampling steps represent the computation necessary to update the value of a
    random variables with a Markov kernel. They are represented as `Op`s with an
    inner graph.

    """

    def __init__(self, inputs: List[TensorVariable], outputs: List[TensorVariable]):
        super().__init__(inputs, outputs, inline=True)


@dataclass(frozen=True)
class Sampler:
    """A class that tracks sampling steps and their parameters."""

    sample_steps: Dict[TensorVariable, TensorVariable]
    """A map between measures and their updated value under the current sampling scheme."""
    updates: Optional[Dict[Variable, TensorVariable]] = field(default_factory=dict)
    """Updates to be passed to `aesara.function`"""
    parameters: Dict[SamplingStep, Tuple[TensorVariable]] = field(default_factory=dict)
    """Parameters needed by the sampling steps."""

    stages: Dict[SamplingStep, List[TensorVariable]] = field(init=False)
    """A list of the sampling stages sorted in scan order."""

    def __post_init__(self):
        # The scan order corresponds to the order in which the sampling
        # steps were found, and thus the order in which elements were
        # inserted in `sample_steps`.
        stages = defaultdict(list)
        for rv, updated_rv in self.sample_steps.items():
            kernel = updated_rv.owner.op
            stages[kernel].append(rv)
        super().__setattr__("stages", stages)



from collections.abc import Mapping
from functools import wraps
from typing import Callable, Dict, Iterable, List, Optional, Sequence, Set, Tuple, Union

from aeppl.rewriting import MeasurableConversionTracker
from aesara.compile.builders import OpFromGraph
from aesara.compile.mode import optdb
from aesara.graph.basic import Apply, Variable, clone_replace, io_toposort
from aesara.graph.features import AlreadyThere, Feature
from aesara.graph.fg import FunctionGraph
from aesara.graph.op import Op
from aesara.graph.rewriting.basic import in2out, node_rewriter
from aesara.graph.rewriting.db import SequenceDB
from aesara.tensor.elemwise import DimShuffle, Elemwise
from aesara.tensor.random.op import RandomVariable
from aesara.tensor.random.utils import RandomStream
from aesara.tensor.rewriting.shape import ShapeFeature
from aesara.tensor.var import TensorVariable
from cons.core import _car
from unification.core import _unify

from aemcmc.utils import remove_constants

SamplerFunctionReturnType = Optional[
    Iterable[Tuple[Variable, Variable, Dict[Variable, Variable]]]
]
SamplerFunctionType = Callable[
    [FunctionGraph, Apply, RandomStream], SamplerFunctionReturnType
]
LocalRewriterReturnType = Optional[Union[Dict[Variable, Variable], Sequence[Variable]]]

sampler_ir_db = SequenceDB()
sampler_ir_db.name = "sampler_ir_db"
sampler_ir_db.register(
    "sampler_canonicalize",
    optdb.query("+canonicalize"),
    "basic",
)

sampler_rewrites_db = SequenceDB()
sampler_rewrites_db.name = "sampler_rewrites_db"

sampler_finder_db = SequenceDB()
sampler_finder_db.name = "sampler_finder_db"

sampler_rewrites_db.register(
    "sampler_finders",
    sampler_finder_db,
    "basic",
    position=0,
)


def construct_ir_fgraph(
    obs_rvs_to_values: Dict[Variable, Variable], clone=True
) -> Tuple[
    FunctionGraph,
    Dict[Variable, Variable],
    Dict[Variable, Variable],
    Dict[Variable, Variable],
]:
    r"""Construct a `FunctionGraph` in measurable IR form for the keys in `obs_rvs_to_values`.

    Returns
    -------
    A `FunctionGraph` of the measurable IR, a copy of `obs_rvs_to_values` containing
    the new, cloned versions of the original variables in `obs_rvs_to_values`, and
    a ``dict`` mapping all the original variables to their cloned values in
    `FunctionGraph`.
    """
    memo = {v: v for v in obs_rvs_to_values.values()}

    rv_outputs = tuple(
        node.outputs[1]
        for node in io_toposort([], list(obs_rvs_to_values.keys()))
        if isinstance(node.op, RandomVariable)
    )

    observed_vars = tuple(obs_rvs_to_values.keys())

    assert all(obs_rv in obs_rvs_to_values for obs_rv in observed_vars)

    fgraph = FunctionGraph(
        outputs=rv_outputs,
        clone=clone,
        memo=memo,
        copy_orphans=False,
        copy_inputs=False,
        features=[ShapeFeature(), MeasurableConversionTracker()],
    )

    # Update `obs_rvs_to_values` so that it uses the new cloned variables
    obs_rvs_to_values = {memo.get(k, k): v for k, v in obs_rvs_to_values.items()}

    sampler_ir_db.query("+basic").rewrite(fgraph)

    new_to_old_rvs = {
        new_rv: old_rv for old_rv, new_rv in zip(rv_outputs, fgraph.outputs)
    }

    return fgraph, obs_rvs_to_values, memo, new_to_old_rvs


class SamplerTracker(Feature):
    """A `Feature` that tracks potential sampler steps in a graph."""

    def __init__(self, srng: RandomStream):
        self.srng: RandomStream = srng
        # Maps variables to a list of tuples that each provide a description of
        # the posterior step, the posterior step's graph/output variable, and
        # any updates generated for the posterior step
        self.rvs_to_samplers: Dict[
            TensorVariable,
            List[Tuple[str, TensorVariable, Optional[Dict[Variable, Variable]]]],
        ] = {}
        self.rvs_seen: Set[TensorVariable] = set()

    def on_attach(self, fgraph: FunctionGraph):
        if hasattr(fgraph, "sampler_mappings"):  # pragma: no cover
            raise AlreadyThere(
                f"{fgraph} already has the `SamplerTracker` feature attached."
            )

        fgraph.sampler_mappings = self


def sampler_finder(tracks: Optional[Sequence[Union[Op, type]]]):
    """Construct a `NodeRewriter` that identifies sample steps.

    This is a decorator that is used as follows:

        @sampler_finder([NormalRV])
        def local_horseshoe_posterior(fgraph, node, srng):
            # Determine if this normal is the root of a Horseshoe
            # prior graph.
            ...
            # If it is, construct the posterior steps for its parameters and
            # return them as a list of tuples like `(rv, posterior_rv, updates)`.
            ...
            return [(lambda_rv, lambda_posterior, None), (tau_rv, tau_posterior, None)]

    """

    def decorator(f: SamplerFunctionType):
        @node_rewriter(tracks)
        @wraps(f)
        def sampler_finder(
            fgraph: FunctionGraph, node: Apply
        ) -> LocalRewriterReturnType:
            sampler_mappings = getattr(fgraph, "sampler_mappings", None)

            # TODO: This assumes that `node` is a `RandomVariable`-generated `Apply` node
            rv_var = node.outputs[1]
            key = (f.__name__, rv_var)

            if sampler_mappings is None or key in sampler_mappings.rvs_seen:
                return None  # pragma: no cover

            srng = sampler_mappings.srng

            rvs_and_posteriors: SamplerFunctionReturnType = f(fgraph, node, srng)

            if not rvs_and_posteriors:
                return None  # pragma: no cover

            for rv, posterior_rv, updates in rvs_and_posteriors:
                sampler_mappings.rvs_to_samplers.setdefault(rv, []).append(
                    (f.__name__, posterior_rv, updates)
                )
            sampler_mappings.rvs_seen.add(key)

            return rv_var.owner.outputs

        return sampler_finder

    return decorator


class SubsumingElemwise(OpFromGraph, Elemwise):
    r"""A class representing an `Elemwise` with `DimShuffle`\ed arguments."""

    def __init__(self, inputs, outputs, *args, **kwargs):
        # TODO: Mock the `Elemwise` interface just enough for our purposes
        self.elemwise_op = outputs[0].owner.op
        self.scalar_op = self.elemwise_op.scalar_op
        self.nfunc_spec = self.elemwise_op.nfunc_spec
        self.inplace_pattern = self.elemwise_op.inplace_pattern
        # self.destroy_map = self.elemwise_op.destroy_map
        self.ufunc = None
        self.nfunc = None

        OpFromGraph.__init__(self, inputs, outputs, *args, **kwargs)

    def make_node(self, *inputs):
        # TODO: We could make sure that the new constant inputs correspond to
        # the originals...
        used_inputs = remove_constants(inputs)

        # The user interface doesn't expect the shared variable inputs of the
        # inner-graph, but, since `Op.make_node` does (and `Op.__call__`
        # dispatches to `Op.make_node`), we need to compensate here
        num_expected_inps = len(self.inner_inputs) - len(self.shared_inputs)

        if len(used_inputs) == num_expected_inps:
            used_inputs = used_inputs + self.shared_inputs

        node = super().make_node(*used_inputs)

        # Remove shared variable inputs.  We aren't going to compute anything
        # with this `Op`, so they're not needed
        real_inputs = node.inputs[: len(node.inputs) - len(self.shared_inputs)]
        return Apply(self, real_inputs, [o.clone() for o in node.outputs])

    def perform(self, *args, **kwargs):
        raise NotImplementedError(  # pragma: no cover
            "This `OpFromGraph` should have been in-line expanded."
        )

    def clone(self):
        res = OpFromGraph.clone(self)
        res.elemwise_op = self.elemwise_op
        res.scalar_op = self.scalar_op
        res.nfunc_spec = self.nfunc_spec
        res.inplace_pattern = self.inplace_pattern
        # res.destroy_map = self.destroy_map
        res.ufunc = self.ufunc
        res.nfunc = self.nfunc
        return res

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return f"{type(self).__name__}{{{self.scalar_op}}}"

    def __eq__(self, other):
        return OpFromGraph.__eq__(self, other)

    def __hash__(self):
        return OpFromGraph.__hash__(self)


def _unify_SubsumingElemwise(u: Elemwise, v: SubsumingElemwise, s: Mapping):
    yield _unify(u, v.elemwise_op, s)


_unify.add(
    (Elemwise, SubsumingElemwise, Mapping),
    lambda u, v, s: _unify_SubsumingElemwise(u, v, s),
)
_unify.add(
    (SubsumingElemwise, Elemwise, Mapping),
    lambda u, v, s: _unify_SubsumingElemwise(v, u, s),
)
_unify.add(
    (SubsumingElemwise, SubsumingElemwise, Mapping),
    lambda u, v, s: _unify(v.elemwise_op, u.elemwise_op, s),
)


def car_SubsumingElemwise(x):
    return type(x.elemwise_op)


_car.add((SubsumingElemwise,), car_SubsumingElemwise)


@node_rewriter([Elemwise])
def local_elemwise_dimshuffle_subsume(fgraph, node):
    r"""This rewrite converts `DimShuffle`s in the `Elemwise` inputs into a single `Op`.

    The replacement rule is

    .. math:

        \frac{
            \operatorname{Elemwise}_{o}\left(
                \operatorname{DimShuffle}_{z_i}(x_i), \dots
            \right)
        }{
            \operatorname{OpFromGraph}_{\operatorname{Elemwise}_{o}\left(
                \operatorname{DimShuffle}_{z_i}(y_i), \dots
            \right)}\left(
                x_i, \dots
            \right)
        }
        //, \quad
        // x_i \text{ is a } \operatorname{RandomVariable}

    where :math:`o` is a scalar `Op`, :math:`z_i` are the `DimShuffle` settings
    for the inputs at index :math:`i`.

    """

    if isinstance(node.op, SubsumingElemwise):
        return None

    new_inputs = []
    subsumed_inputs = []

    out_ndim = node.outputs[0].type.ndim

    found_subsumable_ds = False
    for i in node.inputs:
        if i.owner and isinstance(i.owner.op, DimShuffle):
            # TODO FIXME: Only do this when the `DimShuffle`s are adding
            # broadcastable dimensions.  If they're doing more
            # (e.g. transposing), separate the broadcasting from everything
            # else.
            ds_order = i.owner.op.new_order
            dim_shuffle_input = i.owner.inputs[0]

            ndim_diff = out_ndim - dim_shuffle_input.type.ndim

            # The `DimShuffle`ing added by `Elemwise`
            el_order = ds_order[:ndim_diff]
            # The remaining `DimShuffle`ing that was added by something else
            new_ds_order = ds_order[ndim_diff:]

            # Only consider broadcast dimensions added on the left as
            # having come from `Elemwise.make_node`
            if len(el_order) == 0 or not all(d == "x" for d in el_order):
                # In this case, the necessary broadcast elements were most
                # likely not added by `Elemwise.make_node` (e.g. broadcasts are
                # interspersed with transposes, or there are none at all), so
                # we don't want to mess with them.
                # TODO: We could still subsume some of these `DimShuffle`s,
                # though
                subsumed_inputs.append(i)
                new_inputs.append(i)
                continue

            # if dim_shuffle_input.owner and isinstance(
            #     dim_shuffle_input.owner.op, RandomVariable
            # ):
            found_subsumable_ds = True

            if new_ds_order and not new_ds_order == tuple(range(len(new_ds_order))):
                # The remaining `DimShuffle`ing is substantial, so we need to
                # apply it separately
                new_dim_shuffle_input = dim_shuffle_input.dimshuffle(new_ds_order)
                new_subsumed_input = new_dim_shuffle_input.dimshuffle(
                    el_order + tuple(range(new_dim_shuffle_input.type.ndim))
                )

                subsumed_inputs.append(new_subsumed_input)
                new_inputs.append(new_dim_shuffle_input)
            else:
                subsumed_inputs.append(i)
                new_inputs.append(dim_shuffle_input)

        else:
            subsumed_inputs.append(i)
            new_inputs.append(i)

    if not found_subsumable_ds:
        return None  # pragma: no cover

    assert len(subsumed_inputs) == len(node.inputs)
    new_outputs = node.op.make_node(*subsumed_inputs).outputs

    used_inputs = remove_constants(new_inputs)
    new_op = SubsumingElemwise(used_inputs, new_outputs, inline=True)

    new_out = new_op(*used_inputs)

    return new_out.owner.outputs


sampler_ir_db.register(
    "elemwise_dimshuffle_subsume",
    in2out(local_elemwise_dimshuffle_subsume),
    "basic",
    position=-10,
)


@node_rewriter([Elemwise])
def inline_SubsumingElemwise(fgraph, node):
    op = node.op

    if not isinstance(op, SubsumingElemwise):
        return False

    if not op.is_inline:
        return False  # pragma: no cover

    res = clone_replace(
        op.inner_outputs, {u: v for u, v in zip(op.inner_inputs, node.inputs)}
    )

    return res


expand_subsumptions = in2out(inline_SubsumingElemwise)

# This step undoes `elemwise_dimshuffle_subsume`
sampler_rewrites_db.register(
    "expand_subsumptions",
    expand_subsumptions,
    "basic",
    position="last",
)



from typing import List, Mapping, Optional, Tuple

import aesara
import aesara.tensor as at
from aesara.graph.basic import Variable
from aesara.graph.rewriting.basic import in2out
from aesara.graph.rewriting.db import LocalGroupDB
from aesara.graph.rewriting.unify import eval_if_etuple
from aesara.ifelse import ifelse
from aesara.tensor.math import Dot
from aesara.tensor.random import RandomStream
from aesara.tensor.random.basic import BernoulliRV, NegBinomialRV, NormalRV
from aesara.tensor.var import TensorVariable
from etuples import etuple, etuplize
from unification import unify, var

from aemcmc.dists import (
    multivariate_normal_cong2017,
    multivariate_normal_rue2005,
    polyagamma,
)
from aemcmc.rewriting import sampler_finder, sampler_finder_db
from aemcmc.types import SamplingStep
from aemcmc.utils import remove_constants

gibbs_db = LocalGroupDB(apply_all_rewrites=True)
gibbs_db.name = "gibbs_db"


def normal_regression_overdetermined_posterior(
    srng: RandomStream,
    omega: TensorVariable,
    lmbdatau_inv: TensorVariable,
    X: TensorVariable,
    z: TensorVariable,
) -> TensorVariable:
    """Sample from the posterior of a normal prior and normal observation model.

    This version handles ``X.shape[1] <= X.shape[0]``.

    See `update_beta` for a description of the parameters and return value.

    """
    Q = X.T @ (omega[:, None] * X)
    indices = at.arange(Q.shape[1])
    Q = at.subtensor.set_subtensor(
        Q[indices, indices],
        at.diag(Q) + lmbdatau_inv,
    )
    return multivariate_normal_rue2005(srng, X.T @ (omega * z), Q)


def normal_regression_underdetermined_posterior(
    srng: RandomStream,
    omega: TensorVariable,
    lmbdatau_inv: TensorVariable,
    X: TensorVariable,
    z: TensorVariable,
) -> TensorVariable:
    """Sample from the posterior of a normal prior and normal observation model.

    This version handles ``X.shape[1] > X.shape[0]``.

    See `update_beta` for a description of the parameters and return value.

    """
    return multivariate_normal_cong2017(srng, lmbdatau_inv, omega, X, z)


def normal_regression_posterior(
    srng: RandomStream,
    omega: TensorVariable,
    lmbdatau_inv: TensorVariable,
    X: TensorVariable,
    z: TensorVariable,
) -> TensorVariable:
    r"""Sample from the posterior of a normal prior and normal observation model.

    .. math::

        \begin{align*}
            Z &\sim \operatorname{N}\left( X \beta, \Omega^{-1} \right) \\
            \beta &\sim \operatorname{N}\left( 0, \tau^2 \Lambda \right)
        \end{align*}

    where :math:`X \in \mathbb{R}^{n \times p}`,
    :math:`\Lambda = \operatorname{diag}\left(\lambda^2_1, \dots, \lambda^2_p\right)`, and
    :math:`\Omega^{-1} = \operatorname{diag}\left(\omega_1, \dots, \omega_n\right)`.

    The posterior distribution of :math:`\beta` is given by

    .. math::

        \begin{align*}
            \left( \beta \mid Z = z \right) &\sim
                \operatorname{N}\left( A^{-1} X^{\top} \Omega z, A^{-1} \right) \\
            A &= X^{\top} X + \Lambda^{-1}_{*} \\
            \Lambda_{*} &= \tau^2 \Lambda
        \end{align*}

    This function chooses the best sampler for :math:`\beta \mid z` based on
    the dimensions of :math:`X`.

    Parameters
    ----------
    srng
        The random number generator used to draw samples.
    omega
        The observation model diagonal std. dev. values :math:`\omega_i`.
        In other words, :math:`\operatorname{diag}\left(\Omega\right)`.
    lmbdatau_inv
        The inverse :math:`beta` std. dev. values :math:`\tau^{-1} \lambda^{-1}_i`.
        In other words, :math:`\operatorname{diag}\left(\Lambda^{-1/2}_{*}\right)`.
    X
        Regression matrix :math:`X`.
    z
        Observed values :math:`z \sim Z`.


    Returns
    -------
    A sample from :math:`\beta \mid z`.

    """
    beta_posterior = ifelse(
        X.shape[1] > X.shape[0],
        normal_regression_underdetermined_posterior(srng, omega, lmbdatau_inv, X, z),
        normal_regression_overdetermined_posterior(srng, omega, lmbdatau_inv, X, z),
    )

    return beta_posterior


halfcauchy_1_lv, halfcauchy_2_lv = var(), var()
zero_lv = var()
horseshoe_pattern = etuple(
    etuplize(at.random.normal),
    var(),
    var(),
    var(),
    zero_lv,
    etuple(etuplize(at.mul), halfcauchy_1_lv, halfcauchy_2_lv),
)


def horseshoe_match(graph: TensorVariable) -> Tuple[TensorVariable, TensorVariable]:
    graph_et = etuplize(graph)

    s = unify(graph_et, horseshoe_pattern)
    if s is False:
        raise ValueError("Not a horseshoe prior.")

    halfcauchy_1 = eval_if_etuple(s[halfcauchy_1_lv])
    if halfcauchy_1.owner is None or not isinstance(
        halfcauchy_1.owner.op, type(at.random.halfcauchy)
    ):
        raise ValueError(
            "Not a horseshoe prior. One of the shrinkage parameters "
            + "in your model is not half-Cauchy distributed."
        )

    halfcauchy_2 = eval_if_etuple(s[halfcauchy_2_lv])

    if halfcauchy_2.owner is None or not isinstance(
        halfcauchy_2.owner.op, type(at.random.halfcauchy)
    ):
        raise ValueError(
            "Not a horseshoe prior. One of the shrinkage parameters "
            + "in your model is not half-Cauchy distributed."
        )

    if halfcauchy_1.type.ndim == 0 or all(s == 1 for s in halfcauchy_1.type.shape):
        lmbda_rv = halfcauchy_2
        tau_rv = halfcauchy_1
    elif halfcauchy_2.type.ndim == 0 or all(s == 1 for s in halfcauchy_2.type.shape):
        lmbda_rv = halfcauchy_1
        tau_rv = halfcauchy_2
    else:
        raise ValueError(
            "Not a horseshoe prior. The global shrinkage parameter "
            + "in your model must be one-dimensional."
        )

    return (lmbda_rv, tau_rv)


def horseshoe_posterior(
    srng: RandomStream,
    beta: TensorVariable,
    sigma2: TensorVariable,
    lambda2: TensorVariable,
    tau2: TensorVariable,
) -> Tuple[TensorVariable, TensorVariable]:
    r"""Gibbs kernel to sample from the posterior distributions of the horseshoe prior shrinkage parameters.

    This kernel generates samples from the posterior distribution of the local
    and global shrinkage parameters of a horseshoe prior, respectively :math:`\lambda`
    and :math:`\tau` in the following model:

    .. math::

        \begin{align*}
            \beta_j &\sim \operatorname{N}\left(0, \lambda_j^2 \tau^2 \sigma^2\right) \\
            \lambda_j &\sim \operatorname{C}^{+}(0, 1) \\
            \tau &\sim \operatorname{C}^{+}(0, 1)
        \end{align*}

    The graphs constructed by this function are :math:`\lambda \mid \beta, \tau` and
    :math:`\tau \mid \lambda`, respectively.

    We use the following observations [1]_ to sample from the posterior
    conditional probability of :math:`\tau` and :math:`\lambda`:

    1. The half-Cauchy distribution can be intepreted as a mixture of inverse-gamma
    distributions;
    2. If :math:`Z \sim \operatorname{InvGamma}(1, a)`, :math:`Z \sim 1 / \operatorname{Exp}(a)`.

    Parameters
    ----------
    srng
        The random number generating object to be used during sampling.
    beta
        Regression coefficients.
    sigma2
        Variance of the regression coefficients.
    lambda2
        Square of the local shrinkage parameters.
    tau2
        Square of the global shrinkage parameters.

    Returns
    -------
    Posteriors for :math:`lambda` and :math:`tau`, respectively.

    References
    ----------
    .. [1] Makalic, Enes & Schmidt, Daniel. (2016). High-Dimensional Bayesian
          Regularised Regression with the BayesReg Package.

    """
    lmbda2_inv = at.reciprocal(lambda2)
    tau2_inv = at.reciprocal(tau2)

    upsilon_inv = srng.exponential(1 + lmbda2_inv)
    zeta_inv = srng.exponential(1 + tau2_inv)

    beta2 = beta**2
    lambda2_inv_new = srng.exponential(upsilon_inv + 0.5 * beta2 * tau2_inv / sigma2)
    tau2_inv_new = srng.gamma(
        0.5 * (beta.shape[0] + 1),
        zeta_inv + 0.5 * (beta2 * lambda2_inv_new).sum() / sigma2,
    )

    lambda2_update = at.reciprocal(at.sqrt(lambda2_inv_new))
    tau2_update = at.reciprocal(at.sqrt(tau2_inv_new))

    return lambda2_update, tau2_update


class HorseshoeGibbsKernel(SamplingStep):
    """An `Op` that represents a state update with the FFBS sampler."""


@sampler_finder([NormalRV])
def normal_horseshoe_finder(fgraph, node, srng):
    r"""Find and construct a Gibbs sampler for the normal-Horseshoe model.

    The implementation follows the sampler described in [1]_. It is designed to
    sample efficiently from the following model:

    .. math::

        \begin{align*}
            \beta_i &\sim \operatorname{N}(0, \lambda_i^2 \tau^2) \\
            \lambda_i &\sim \operatorname{C}^{+}(0, 1) \\
            \tau &\sim \operatorname{C}^{+}(0, 1)
        \end{align*}

    References
    ----------
    .. [1] Makalic, Enes & Schmidt, Daniel. (2015). A Simple Sampler for the
          Horseshoe Estimator. 10.1109/LSP.2015.2503725.

    """

    rv_var = node.outputs[1]

    try:
        lambda_rv, tau_rv = horseshoe_match(node.outputs[1])
    except ValueError:  # pragma: no cover
        return None

    tau2 = tau_rv**2
    lambda2 = lambda_rv**2
    lambda_posterior, tau_posterior = horseshoe_posterior(
        srng, rv_var, at.as_tensor(1.0), lambda2, tau2
    )

    # Build an `Op` for the sampling kernel
    outputs = [lambda_posterior, tau_posterior]
    inputs = remove_constants([rv_var, lambda2, tau2])
    gibbs = HorseshoeGibbsKernel(inputs, outputs)

    lambda_posterior, tau_posterior = gibbs(*inputs)
    lambda_posterior.name = f"{lambda_rv.name or 'lambda'}_posterior"
    tau_posterior.name = f"{tau_rv.name or 'tau'}_posterior"

    return [(lambda_rv, lambda_posterior, None), (tau_rv, tau_posterior, None)]


gibbs_db.register("normal_horseshoe", normal_horseshoe_finder, "basic")


X_lv = var()
beta_lv = var()
neg_one_lv = var()

sigmoid_dot_pattern = etuple(
    etuplize(at.sigmoid),
    etuple(etuplize(at.mul), neg_one_lv, etuple(etuple(Dot), X_lv, beta_lv)),
)

a_lv = var()
b_lv = var()
gamma_pattern = etuple(etuplize(at.random.gamma), var(), var(), var(), a_lv, b_lv)


def gamma_match(graph: TensorVariable) -> Tuple[TensorVariable, TensorVariable]:
    graph_et = etuplize(graph)
    s = unify(graph_et, gamma_pattern)
    if s is False:
        raise ValueError("Not a gamma prior.")

    a = eval_if_etuple(s[a_lv])
    b = eval_if_etuple(s[b_lv])

    return a, b


h_lv = var()
nbinom_sigmoid_dot_pattern = etuple(
    etuplize(at.random.nbinom), var(), var(), var(), h_lv, sigmoid_dot_pattern
)


def nbinom_sigmoid_dot_match(
    graph: TensorVariable,
) -> Tuple[TensorVariable, TensorVariable, TensorVariable]:
    graph_et = etuplize(graph)
    s = unify(graph_et, nbinom_sigmoid_dot_pattern)
    if s is False:
        raise ValueError("Not a negative binomial regression.")

    if all(s[neg_one_lv].data != -1):
        raise ValueError(
            "Not a negative binomial regression. The argument to "
            + "the sigmoid must be minus the dot product."
        )

    h = eval_if_etuple(s[h_lv])
    beta_rv = eval_if_etuple(s[beta_lv])
    X = eval_if_etuple(s[X_lv])

    return X, h, beta_rv


def sample_CRT(
    srng: RandomStream, y: TensorVariable, h: TensorVariable
) -> Tuple[TensorVariable, Mapping[Variable, Variable]]:
    r"""Sample a Chinese Restaurant Process value: :math:`l \sim \operatorname{CRT}(y, h)`.

    Sampling is performed according to the following:

    .. math::

        \begin{gather*}
            l = \sum_{n=1}^{y} b_n, \quad
            b_n \sim \operatorname{Bern}\left(\frac{h}{n - 1 + h}\right)
        \end{gather*}

    References
    ----------
    .. [1] Zhou, Mingyuan, and Lawrence Carin. 2012. “Augment-and-Conquer Negative Binomial Processes.” Advances in Neural Information Processing Systems 25.

    """

    def single_sample_CRT(y_i: TensorVariable, h: TensorVariable):
        n_i = at.arange(2, y_i + 1)
        return at.switch(y_i < 1, 0, 1 + srng.bernoulli(h / (n_i - 1 + h)).sum())

    res, updates = aesara.scan(
        single_sample_CRT,
        sequences=[y.ravel()],
        non_sequences=[h],
        strict=True,
    )
    res = res.reshape(y.shape)
    res.name = "CRT sample"

    return res, updates


def nbinom_dispersion_posterior(
    srng: RandomStream,
    h: TensorVariable,
    p: TensorVariable,
    a: TensorVariable,
    b: TensorVariable,
    y: TensorVariable,
) -> Tuple[TensorVariable, Mapping[Variable, Variable]]:
    r"""Sample the conditional posterior for the dispersion parameter under a negative-binomial and gamma prior.

    In other words, this draws a sample from :math:`h \mid Y = y` per

    .. math::

        \begin{align*}
            Y_i &\sim \operatorname{NB}(h, p_i) \\
            h &\sim \operatorname{Gamma}(a, b)
        \end{align*}

    where :math:`\operatorname{NB}` is a negative-binomial distribution.

    The conditional posterior sample step is derived from the following decomposition:

    .. math::
        \begin{gather*}
            Y_i = \sum_{j=1}^{l_i} u_{i j}, \quad u_{i j} \sim \operatorname{Log}(p_i), \quad
            l_i \sim \operatorname{Pois}\left(-h \log(1 - p_i)\right)
        \end{gather*}

    where :math:`\operatorname{Log}` is the logarithmic distribution.  Under a
    gamma prior, :math:`h` is conjugate to :math:`l`.  We draw samples from
    :math:`l` according to :math:`l \sim \operatorname{CRT(y, h)}`, where
    :math:`y` is a sample from :math:`y \sim Y`.

    The resulting posterior is

    .. math::

        \begin{gather*}
            \left(h \mid Y = y\right) \sim \operatorname{Gamma}\left(a + \sum_{i=1}^N l_i, \frac{1}{1/b + \sum_{i=1}^N \log(1 - p_i)} \right)
        \end{gather*}

    Parameters
    ----------
    srng
        The random number generator from which samples are drawn.
    h
        The  value of :math:`h`.
    p
        The success probability parameter in the negative-binomial distribution of :math:`Y`.
    a
        The shape parameter in the :math:`\operatorname{Gamma}` prior on :math:`h`.
    b
        The rate parameter in the :math:`\operatorname{Gamma}` prior on :math:`h`.
    y
        A sample from :math:`Y`.

    Returns
    -------
    A sample from the posterior :math:`h \mid y`.

    References
    ----------
    .. [1] Zhou, Mingyuan, Lingbo Li, David Dunson, and Lawrence Carin. 2012.
        “Lognormal and Gamma Mixed Negative Binomial Regression.”
        Proceedings of the International Conference on Machine Learning.
        2012: 1343–50. https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4180062/.

    """
    Ls, updates = sample_CRT(srng, y, h)
    L_sum = Ls.sum(axis=-1)
    h_posterior = srng.gamma(
        a + L_sum, at.reciprocal(b) - at.sum(at.log(1 - p), axis=-1)
    )

    return h_posterior, updates


def nbinom_normal_posterior(srng, beta, beta_std, X, h, y):
    r"""Produce a Gibbs sample step for a negative binomial logistic-link regression with a normal prior.

    The implementation follows the sampler described in [1]_. It is designed to
    sample efficiently from the following negative binomial regression model:

    .. math::

        \begin{align*}
            Y &\sim \operatorname{NB}\left(h, p\right) \\
            p &= \frac{\exp(\psi)}{1 + \exp(\psi)} \\
            \psi &= X^\top \beta \\
            \beta &\sim \operatorname{N}(0, \lambda^2) \\
        \end{align*}

    where :math:`\operatorname{NB}` is a negative-binomial distribution.

    Parameters
    ----------
    srng
        The random number generator from which samples are drawn.
    beta
        The current/previous value of the regression parameter :math:`beta`.
    beta_std
        The std. dev. of the regression parameter :math:`beta`.
    X
        The regression matrix.
    h
        The :math:`h` parameter in the negative-binomial distribution of :math:`Y`.
    y
        A sample from the observation distribution :math:`y \sim Y`.

    Returns
    -------
    A sample from the posterior :math:`\beta \mid y`.

    Notes
    -----
    The :math:`z` expression in Section 2.2 of [1]_ seems to
    omit division by the Polya-Gamma auxilliary variables whereas [2]_ and [3]_
    explicitly include it. We found that including the division results in
    accurate posterior samples for the regression coefficients. It is also
    worth noting that the :math:`\sigma^2` parameter is not sampled directly
    in the negative binomial regression problem and thus set to 1 [2]_.

    References
    ----------
    .. [1] Makalic, Enes & Schmidt, Daniel. (2015). A Simple Sampler for the
          Horseshoe Estimator. 10.1109/LSP.2015.2503725.
    .. [2] Makalic, Enes & Schmidt, Daniel. (2016). High-Dimensional Bayesian
          Regularised Regression with the BayesReg Package.
    .. [3] Neelon, Brian. (2019). Bayesian Zero-Inflated Negative Binomial
          Regression Based on Pólya-Gamma Mixtures. Bayesian Anal.
          2019 September ; 14(3): 829–855. doi:10.1214/18-ba1132.

    """

    # This effectively introduces a new term, `w`, to the model.
    # TODO: We could/should generate a graph that uses this scale-mixture
    # "expanded" form and find/create the posteriors from there
    w = srng.gen(polyagamma, y + h, X @ beta)
    z = 0.5 * (y - h) / w

    tau_beta = at.reciprocal(beta_std)

    beta_posterior = normal_regression_posterior(srng, w, tau_beta, X, z)

    return beta_posterior


class NBRegressionGibbsKernel(SamplingStep):
    """An `Op` that represents the update of the regression parameter of
    a negative binomial regression.

    """

    default_output = 0


class DispersionGibbsKernel(SamplingStep):
    """An `Op` that represents the state update for the dispersion parameter
    of a negative binomial in a negative binomial regression.

    """

    default_output = 0


@sampler_finder([NegBinomialRV])
def nbinom_logistic_finder(fgraph, node, srng):
    r"""Find and construct a Gibbs sampler for a negative-binomial logistic-link regression.

    The implementation follows the sampler described in `nbinom_normal_posterior`. It is designed to
    sample efficiently from the following negative binomial regression model:

    .. math::

        \begin{align*}
            Y &\sim \operatorname{NB}\left(h, p\right) \\
            p &= \frac{\exp(\psi)}{1 + \exp(\psi)} \\
            \psi &= X^\top \beta \\
            \beta_j &\sim \operatorname{N}(0, \lambda_j^2) \\
            h \sim \operatorname{Gamma}\left(a, b\right)
        \end{align*}

    If :math:`h` doesn't take the above form, a sampler is produced with steps
    for all the other terms; otherwise, sampling for :math:`h` is performed
    in accordance with [1]_.

    References
    ----------
    .. [1] Zhou, Mingyuan, Lingbo Li, David Dunson, and Lawrence Carin. 2012.
          Lognormal and Gamma Mixed Negative Binomial Regression.
          Proceedings of the International Conference on Machine Learning.
          2012: 1343–50. https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4180062/.

    """

    y = node.outputs[1]

    try:
        X, h, beta_rv = nbinom_sigmoid_dot_match(node.outputs[1])
    except ValueError:  # pragma: no cover
        return None

    beta_std = beta_rv.owner.inputs[4]
    beta_posterior = nbinom_normal_posterior(srng, beta_rv, beta_std, X, h, y)

    # Build the `Op` corresponding to the sampling step
    outputs = [beta_posterior]
    inputs = remove_constants([beta_rv, beta_std, X, h, y])
    gibbs = NBRegressionGibbsKernel(inputs, outputs)

    beta_posterior = gibbs(*inputs)
    beta_posterior.name = f"{beta_rv.name or 'beta'}_posterior"

    res: List[
        Tuple[TensorVariable, TensorVariable, Optional[Mapping[Variable, Variable]]]
    ] = [(beta_rv, beta_posterior, None)]

    # TODO: Should this be in a separate rewriter?
    try:
        a, b = gamma_match(h)
    except ValueError:  # pragma: no cover
        return res

    p = at.sigmoid(-X @ beta_posterior)

    h_posterior, updates = nbinom_dispersion_posterior(srng, h, p, a, b, y)

    # Build the `Op` corresponding to the sampling step
    update_outputs = [h_posterior.owner.inputs[0].default_update]
    update_outputs.extend(updates.values())

    outputs = [h_posterior] + update_outputs
    inputs = remove_constants([h, p, a, b, y])
    gibbs = DispersionGibbsKernel(inputs, outputs)

    h_posterior = gibbs(*inputs)
    h_posterior.name = f"{h.name or 'h'}_posterior"

    updates_offset = len(inputs)
    updates = {
        h_posterior.owner.inputs[updates_offset]: h_posterior.owner.outputs[1],
        h_posterior.owner.inputs[updates_offset + 1]: h_posterior.owner.outputs[2],
    }

    res.append((h, h_posterior, updates))

    return res


gibbs_db.register("nbinom_logistic_regression", nbinom_logistic_finder, "basic")


bernoulli_sigmoid_dot_pattern = etuple(
    etuplize(at.random.bernoulli), var(), var(), var(), sigmoid_dot_pattern
)


def bern_sigmoid_dot_match(
    graph: TensorVariable,
) -> Tuple[TensorVariable, TensorVariable]:
    graph_et = etuplize(graph)

    s = unify(graph_et, bernoulli_sigmoid_dot_pattern)
    if s is False:
        raise ValueError("Not a Bernoulli regression.")

    if all(s[neg_one_lv].data != -1):
        raise ValueError(
            "Not a Bernoulli regression. The argument to the sigmoid "
            + "must be minus the dot product."
        )

    beta_rv = eval_if_etuple(s[beta_lv])
    X = eval_if_etuple(s[X_lv])

    return X, beta_rv


def bern_normal_posterior(
    srng: RandomStream,
    beta: TensorVariable,
    beta_std: TensorVariable,
    X: TensorVariable,
    y: TensorVariable,
) -> Tuple[TensorVariable, TensorVariable, TensorVariable]:
    r"""Produce a Gibbs sample step for a bernoulli logistic-link regression with a normal prior.

    The implementation follows the sampler described in [1]_. It is designed to
    sample efficiently from the following binary logistic regression model:

    .. math::

        \begin{align*}
            Y &\sim \operatorname{Bern}\left( p \right) \\
            p &= \frac{1}{1 + \exp\left( -X^\top \beta\right)} \\
            \beta_j &\sim \operatorname{N}\left( 0, \lambda_j^2 \right)
        \end{align*}


    Parameters
    ----------
    beta
        The current/previous value of the regression parameter :math:`beta`.
    beta_std
        The std. dev. of the regression parameter :math:`beta`.
    X
        The regression matrix.
    y
        A sample from the observation distribution :math:`y \sim Y`.

    Returns
    -------
    A sample from :math:`\beta \mid y`.

    References
    ----------
    .. [1] Makalic, Enes & Schmidt, Daniel. (2016). High-Dimensional Bayesian
          Regularised Regression with the BayesReg Package.

    """
    w = srng.gen(polyagamma, 1, X @ beta)
    z = (y - 0.5) / w

    tau_beta = at.reciprocal(beta_std)

    beta_posterior = normal_regression_posterior(srng, w, tau_beta, X, z)

    return beta_posterior


class BernoulliRegressionGibbsKernel(SamplingStep):
    """An `Op` that represents the update of the regression parameter of
    a Bernoulli regression.

    """

    default_output = 0


@sampler_finder([BernoulliRV])
def bern_logistic_finder(fgraph, node, srng):
    r"""Find and construct a Gibbs sampler for a negative binomial logistic-link regression."""

    y = node.outputs[1]

    try:
        X, beta_rv = bern_sigmoid_dot_match(node.outputs[1])
    except ValueError:  # pragma: no cover
        return None

    beta_std = beta_rv.owner.inputs[4]
    beta_posterior = bern_normal_posterior(srng, beta_rv, beta_std, X, y)

    # Build the `Op` corresponding to the sampling step
    outputs = [beta_posterior]
    inputs = remove_constants([beta_rv, beta_std, X, y])
    gibbs = BernoulliRegressionGibbsKernel(inputs, outputs)

    beta_posterior: TensorVariable = gibbs(*inputs)  # type: ignore
    beta_posterior.name = f"{beta_rv.name or 'beta'}_posterior"  # type: ignore

    return [(beta_rv, beta_posterior, None)]


gibbs_db.register("bern_logistic_finder", bern_logistic_finder, "basic")

sampler_finder_db.register(
    "gibbs_db", in2out(gibbs_db.query("+basic"), name="gibbs"), "basic"
)



import aesara.tensor as at
import pytest
from aesara.graph.fg import FunctionGraph
from aesara.graph.kanren import KanrenRelationSub

from aemcmc.transforms import invgamma_exponential, location_scale_transform


def test_normal_scale_loc_transform_lift():
    srng = at.random.RandomStream(0)
    mu_rv = srng.halfnormal(1.0)
    sigma_rv = srng.halfcauchy(1)
    Y_rv = srng.normal(mu_rv, sigma_rv)

    fgraph = FunctionGraph(outputs=[Y_rv], clone=False)
    res = KanrenRelationSub(location_scale_transform).transform(
        fgraph, fgraph.outputs[0].owner
    )[0]

    # Make sure that Y_rv gets replaced with an addition
    assert res.owner.op == at.add
    lhs = res.owner.inputs[0]
    assert isinstance(lhs.owner.op, type(at.random.halfnormal))
    rhs = res.owner.inputs[1]
    assert rhs.owner.op == at.mul
    assert isinstance(rhs.owner.inputs[0].owner.op, type(at.random.halfcauchy))
    assert isinstance(rhs.owner.inputs[1].owner.op, type(at.random.normal))


@pytest.mark.xfail(
    reason="Op.__call__ does not dispatch to Op.make_node for some RandomVariable and etuple evaluation returns an error"
)
def test_normal_scale_loc_transform_sink():
    srng = at.random.RandomStream(0)
    mu_rv = srng.halfnormal(1.0)
    sigma_rv = srng.halfcauchy(1)
    std_normal_rv = srng.normal(0, 1)
    Y_at = mu_rv + sigma_rv * std_normal_rv

    fgraph = FunctionGraph(outputs=[Y_at], clone=False)
    res = KanrenRelationSub(lambda x, y: location_scale_transform(y, x)).transform(
        fgraph, fgraph.outputs[0].owner
    )[0]

    assert isinstance(res.owner.op, type(at.random.normal))


def test_invgamma_to_exp():
    srng = at.random.RandomStream(0)
    c_at = at.scalar()
    X_rv = srng.invgamma(1.0, c_at)

    fgraph = FunctionGraph(outputs=[X_rv], clone=False)
    res = KanrenRelationSub(invgamma_exponential).transform(
        fgraph, fgraph.outputs[0].owner
    )[0]

    Y_rv = 1.0 / srng.exponential(c_at)

    assert res.owner.op == Y_rv.owner.op
    assert isinstance(res.owner.inputs[1].owner.op, type(Y_rv.owner.inputs[1].owner.op))
    assert res.owner.inputs[1].owner.inputs[-1] == c_at


@pytest.mark.xfail(
    reason="Op.__call__ does not dispatch to Op.make_node for some RandomVariable and etuple evaluation returns an error"
)
def test_invgamma_from_exp():
    srng = at.random.RandomStream(0)
    c_at = at.scalar()
    X_rv = 1.0 / srng.exponential(c_at)

    fgraph = FunctionGraph(outputs=[X_rv], clone=False)
    res = KanrenRelationSub(lambda x, y: invgamma_exponential(y, x)).transform(
        fgraph, fgraph.outputs[0].owner
    )[0]

    Y_rv = srng.invgamma(1.0, c_at)

    assert isinstance(res.owner.op, type(Y_rv.owner.op))
    assert res.owner.inputs[-1] == c_at



from typing import Callable, Dict, Tuple

import aesara
import aesara.tensor as at
from aehmc import nuts as aehmc_nuts
from aehmc.utils import RaveledParamsMap
from aeppl import joint_logprob
from aeppl.transforms import (
    RVTransform,
    TransformValuesRewrite,
    _default_transformed_rv,
)
from aesara import config
from aesara.compile.sharedvalue import SharedVariable
from aesara.graph.basic import Apply, graph_inputs
from aesara.graph.type import Constant
from aesara.tensor.random import RandomStream
from aesara.tensor.random.op import RandomVariable
from aesara.tensor.var import TensorVariable

from aemcmc.types import SamplingStep

NUTSStateType = Tuple[TensorVariable, TensorVariable, TensorVariable]
NUTSKernelType = Callable[
    [NUTSStateType],
    Tuple[
        Tuple[
            NUTSStateType,
            Dict[TensorVariable, TensorVariable],
            Dict[TensorVariable, TensorVariable],
        ],
        Dict,
    ],
]


class NUTSKernel(SamplingStep):
    """An `Op` that represents the update of one or many random variables
    with the NUTS sampling algorithm.

    """


def step(
    srng: RandomStream,
    to_sample_rvs: Dict[RandomVariable, TensorVariable],
    realized_rvs_to_values: Dict[RandomVariable, TensorVariable],
) -> Tuple[
    Dict[RandomVariable, TensorVariable], Dict, Tuple[TensorVariable, TensorVariable]
]:
    """Build a NUTS sampling step and its initial state.

    This sampling step works with variables in their original space, to create
    the sampling step we thus need to:

    1. Create the initial value for the variables to sample;
    2. Create a log-density graph that works in the transformed space, and
       build a NUTS kernel that uses this graph;
    3. Apply the default transformations to the initial values;
    4. Apply the NUTS kernel to the transformed initial values;
    5. Apply the backward transformation to the updated values.

    Parameters
    ----------
    rvs_to_samples
        A dictionary that maps the random variables whose posterior
        distribution we wish to sample from to their initial values.
    realized_rvs_to_values
        A dictionary that maps the random variables not sampled by NUTS to their
        realized value. These variables can either correpond to observations, or
        to variables whose value is set by a different sampler.

    Returns
    -------
    A NUTS sampling step for each random variable, their initial values, the
    shared variable updates and the NUTS parameters.

    """

    # Get the initial values for the random variables that are assigned this
    # sampling step.
    initial_values = to_sample_rvs.values()

    # Algorithms in the HMC family can more easily explore the posterior distribution
    # when the support of each random variable's distribution is unconstrained. Get
    # the default transform that corresponds to each random variable.
    transforms = {rv: get_transform(rv) for rv in to_sample_rvs}
    transformed_values = [
        transform_forward(rv, vv, transforms[rv])
        for rv, vv in zip(to_sample_rvs, initial_values)
    ]

    # Build the graph for the joint log-density of the model, setting the value
    # of the realized variables. The placeholder nodes are defined in the
    # transformed space and will be replaced by their actual value when building
    # the NUTS kernel.
    logprob, placeholder_values = joint_logprob(
        *to_sample_rvs.keys(),
        realized=realized_rvs_to_values,
        extra_rewrites=TransformValuesRewrite(transforms),
    )

    # Algorithms in `AeHMC` work with flat arrays so we need to ravel parameter
    # values to use them as an input to the NUTS kernel. The following creates
    # an object that can ravel the `transformed_values` and can unravel
    # flattened values in a dictionary that maps placeholder values to the
    # unraveled values.
    rp_map = RaveledParamsMap(transformed_values)
    rp_map.ref_params = placeholder_values

    # Within the NUTS kernel we unravel the flat position and replace the
    # placeholder values by their current value in the logprob graph.
    def logprob_fn(q):
        unraveled_q = rp_map.unravel_params(q)
        memo = aesara.graph.basic.clone_get_equiv(
            [], [logprob], copy_inputs=False, copy_orphans=False, memo=unraveled_q
        )
        return memo[logprob]

    # Build the NUTS kernel and initialize the state
    nuts_kernel = aehmc_nuts.new_kernel(srng, logprob_fn)

    initial_q = rp_map.ravel_params(transformed_values)
    initial_state = aehmc_nuts.new_state(initial_q, logprob_fn)

    # Initialize the parameter values
    step_size = at.scalar("step_size", dtype=config.floatX)
    inverse_mass_matrix = at.tensor(
        name="inverse_mass_matrix", shape=initial_q.type.shape, dtype=config.floatX
    )

    # Apply the NUTS kernel to the initial state, unravel and transform the updated
    # values back to the original space.
    (new_q, *_), updates = nuts_kernel(*initial_state, step_size, inverse_mass_matrix)
    transformed_params = rp_map.unravel_params(new_q)
    results = {
        rv: transform_backward(rv, transformed_params[pv], transforms[rv])
        for rv, pv in zip(to_sample_rvs, placeholder_values)
    }

    return (
        results,
        updates,
        (step_size, inverse_mass_matrix),
    )


def construct_sampler(
    srng: RandomStream,
    to_sample_rvs: Dict[RandomVariable, TensorVariable],
    realized_rvs_to_values: Dict[RandomVariable, TensorVariable],
) -> Tuple[Dict[RandomVariable, TensorVariable], Dict, Dict[Apply, TensorVariable]]:
    results, updates, parameters = step(srng, to_sample_rvs, realized_rvs_to_values)

    # Build an `Op` that represents the NUTS sampling step
    update_outputs = list(updates.values())
    outputs = list(results.values()) + update_outputs
    inputs = [
        var_in
        for var_in in graph_inputs(outputs)
        if not isinstance(var_in, Constant) and not isinstance(var_in, SharedVariable)
    ]
    nuts_op = NUTSKernel(inputs, outputs)

    posterior = nuts_op(*inputs)
    results = {rv: posterior[i] for i, rv in enumerate(to_sample_rvs)}

    updates_input = posterior[0].owner.inputs[len(inputs) :]
    updates_output = posterior[len(results) :]
    updates = {
        updates_input[i]: update_out for i, update_out in enumerate(updates_output)
    }

    return results, updates, {nuts_op: parameters}


def get_transform(rv: TensorVariable):
    """Get the default transform associated with the random variable."""
    transform = _default_transformed_rv(rv.owner.op, rv.owner)
    if transform:
        return transform.op.transform
    else:
        return None


def transform_forward(rv: TensorVariable, vv: TensorVariable, transform: RVTransform):
    """Push variables to the transformed space."""
    if transform:
        res = transform.forward(vv, *rv.owner.inputs)
        if vv.name:
            res.name = f"{vv.name}_trans"
        return res
    else:
        return vv


def transform_backward(rv: TensorVariable, vv: TensorVariable, transform: RVTransform):
    """Pull variables back from the transformed space."""
    if transform:
        res = transform.backward(vv, *rv.owner.inputs)
        return res
    else:
        return vv



import aesara
import aesara.tensor as at
import numpy as np
import pytest
from aesara.graph.basic import graph_inputs, io_toposort
from aesara.tensor.random import RandomStream
from aesara.tensor.random.basic import BetaRV, GammaRV
from scipy.linalg import toeplitz

from aemcmc.basic import construct_sampler
from aemcmc.gibbs import (
    DispersionGibbsKernel,
    HorseshoeGibbsKernel,
    NBRegressionGibbsKernel,
)
from aemcmc.rewriting import SubsumingElemwise


def test_closed_form_posterior_beta_binomial():
    srng = RandomStream(0)

    alpha_tt = at.scalar("alpha")
    beta_tt = at.scalar("beta")
    p_rv = srng.beta(alpha_tt, beta_tt, name="p")

    n_tt = at.iscalar("n")
    Y_rv = srng.binomial(n_tt, p_rv, name="Y")

    y_vv = Y_rv.clone()
    y_vv.name = "y"

    sampler, initial_values = construct_sampler({Y_rv: y_vv}, srng)

    p_posterior_step = sampler.sample_steps[p_rv]
    assert len(sampler.parameters) == 0
    assert len(sampler.stages) == 1
    assert isinstance(p_posterior_step.owner.op, BetaRV)


def test_closed_form_posterior_gamma_poisson():
    srng = RandomStream(0)

    alpha_tt = at.scalar("alpha")
    beta_tt = at.scalar("beta")
    l_rv = srng.gamma(alpha_tt, beta_tt, name="p")

    Y_rv = srng.poisson(l_rv, name="Y")

    y_vv = Y_rv.clone()
    y_vv.name = "y"

    sampler, initial_values = construct_sampler({Y_rv: y_vv}, srng)

    p_posterior_step = sampler.sample_steps[l_rv]
    assert len(sampler.parameters) == 0
    assert len(sampler.stages) == 1
    assert isinstance(p_posterior_step.owner.op, GammaRV)


def test_closed_form_posterior_beta_nbinom():
    srng = RandomStream(0)

    alpha_tt = at.scalar("alpha")
    beta_tt = at.scalar("beta")

    p_rv = srng.beta(alpha_tt, beta_tt, name="p")

    n_tt = at.scalar("n")
    Y_rv = srng.negative_binomial(n_tt, p_rv, name="Y")

    y_vv = Y_rv.clone()
    y_vv.name = "y"

    sampler, initial_values = construct_sampler({Y_rv: y_vv}, srng)

    p_posterior_step = sampler.sample_steps[p_rv]
    assert len(sampler.parameters) == 0
    assert len(sampler.stages) == 1
    assert isinstance(p_posterior_step.owner.op, BetaRV)


@pytest.mark.parametrize("size", [1, (1,), (2, 3)])
def test_nuts_sampler_single_variable(size):
    """We make sure that the NUTS sampler compiles and updates the chains for
    different sizes of the random variable.

    """
    srng = RandomStream(0)

    tau_rv = srng.halfcauchy(0, 1, size=size, name="tau")
    Y_rv = srng.halfcauchy(0, tau_rv, name="Y")

    y_vv = Y_rv.clone()
    y_vv.name = "y"

    sampler, initial_values = construct_sampler({Y_rv: y_vv}, srng)
    assert len(sampler.sample_steps) == 1
    assert len(sampler.stages) == 1

    tau_post_step = sampler.sample_steps[tau_rv]
    nuts = tau_post_step.owner.op
    assert y_vv in graph_inputs([tau_post_step])
    assert len(sampler.parameters[nuts]) == 2

    inputs = [
        initial_values[tau_rv],
        y_vv,
        sampler.parameters[nuts][0],
        sampler.parameters[nuts][1],
    ]
    output = tau_post_step
    sample_step = aesara.function(inputs, output)

    tau_val = np.ones(shape=size)
    y_val = np.ones(shape=size)
    step_size = 1e-1
    inverse_mass_matrix = np.ones(shape=size).flatten()
    res = sample_step(tau_val, y_val, step_size, inverse_mass_matrix)
    with pytest.raises(AssertionError):
        np.testing.assert_equal(tau_val, res)


def test_nuts_with_closed_form():
    """Make sure that the NUTS sampler works in combination with closed-form posteriors."""
    srng = RandomStream(0)

    alpha_tt = at.scalar("alpha")
    beta_rv = srng.halfnormal(1.0, name="beta")
    l_rv = srng.gamma(alpha_tt, beta_rv, name="p")

    Y_rv = srng.poisson(l_rv, name="Y")

    y_vv = Y_rv.clone()
    y_vv.name = "y"

    sampler, initial_values = construct_sampler({Y_rv: y_vv}, srng)

    assert len(sampler.stages) == 2

    p_posterior_step = sampler.sample_steps[l_rv]
    assert y_vv in graph_inputs([p_posterior_step])
    assert len(initial_values) == 2
    assert isinstance(p_posterior_step.owner.op, GammaRV)

    assert beta_rv in sampler.sample_steps


def test_create_gibbs():
    srng = RandomStream(0)

    X = at.matrix("X")

    # Horseshoe `beta_rv`
    tau_rv = srng.halfcauchy(0, 1, name="tau")
    lmbda_rv = srng.halfcauchy(0, 1, size=X.shape[1], name="lambda")
    beta_rv = srng.normal(0, lmbda_rv * tau_rv, size=X.shape[1], name="beta")

    a = at.scalar("a")
    b = at.scalar("b")
    h_rv = srng.gamma(a, b, name="h")

    # Negative-binomial regression
    eta = X @ beta_rv
    p = at.sigmoid(-eta)
    Y_rv = srng.nbinom(h_rv, p, name="Y")

    y_vv = Y_rv.clone()
    y_vv.name = "y"

    sample_vars = [tau_rv, lmbda_rv, beta_rv, h_rv]

    sampler, initial_values = construct_sampler({Y_rv: y_vv}, srng)

    assert len(sampler.sample_steps) == 4
    assert len(sampler.stages) == 3
    assert sampler.updates

    tau_post_step = sampler.sample_steps[tau_rv]
    # These are *very* rough checks of the resulting graphs
    assert isinstance(tau_post_step.owner.op, HorseshoeGibbsKernel)

    lmbda_post_step = sampler.sample_steps[lmbda_rv]
    assert isinstance(lmbda_post_step.owner.op, HorseshoeGibbsKernel)

    beta_post_step = sampler.sample_steps[beta_rv]
    assert isinstance(beta_post_step.owner.op, NBRegressionGibbsKernel)

    h_post_step = sampler.sample_steps[h_rv]
    assert isinstance(h_post_step.owner.op, DispersionGibbsKernel)

    inputs = [X, a, b, y_vv] + [initial_values[rv] for rv in sample_vars]
    outputs = [sampler.sample_steps[rv] for rv in sample_vars]

    subsuming_elemwises = [
        n for n in io_toposort([], outputs) if isinstance(n.op, SubsumingElemwise)
    ]
    assert not any(subsuming_elemwises)

    sample_step = aesara.function(
        inputs,
        outputs,
        updates=sampler.updates,
        on_unused_input="ignore",
    )

    rng = np.random.default_rng(2309)

    N = 100
    M = 10
    S = toeplitz(0.5 ** np.arange(M))
    X_val = rng.multivariate_normal(np.zeros(M), S, size=N)

    a_val, b_val = 100.0, 1.0
    beta_true = np.array([2, 0.02, 0.2, 0.1, 1] + [0.0] * (M - 5))
    tau_val, lmbda_val, h_val = 1.0, np.ones(M), 10.0

    y_fn = aesara.function([X, a, b, beta_rv], Y_rv)
    y_val = y_fn(X_val, a_val, b_val, beta_true)

    beta_pst_vals = []
    tau_pst_val, lmbda_pst_val, beta_pst_val, h_pst_val = (
        tau_val,
        lmbda_val,
        np.zeros(M),
        h_val,
    )
    for i in range(100):
        tau_pst_val, lmbda_pst_val, beta_pst_val, h_pst_val = sample_step(
            X_val,
            a_val,
            b_val,
            y_val,
            tau_pst_val,
            lmbda_pst_val,
            beta_pst_val,
            h_pst_val,
        )
        beta_pst_vals += [beta_pst_val]

    beta_pst_mean = np.mean(beta_pst_vals, axis=0)
    assert np.allclose(beta_pst_mean, beta_true, atol=1e-1, rtol=1e-1)



from . import _version

__version__ = _version.get_versions()["version"]


from aemcmc.basic import construct_sampler

# isort: off
# Register rewrite databases
import aemcmc.conjugates
import aemcmc.gibbs

# isort: on



#!/usr/bin/env python
import os
from os.path import exists

from setuptools import setup

import versioneer

NAME = "aemcmc"

# Handle builds of nightly release
if "BUILD_AEMCMC_NIGHTLY" in os.environ:
    nightly = True
    NAME += "-nightly"

    from versioneer import get_versions as original_get_versions

    def get_versions():
        from datetime import datetime, timezone

        suffix = datetime.now(timezone.utc).strftime(r".dev%Y%m%d")
        versions = original_get_versions()
        versions["version"] = versions["version"].split("+")[0] + suffix
        return versions

    versioneer.get_versions = get_versions


setup(
    name=NAME,
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="Miscellaneous MCMC samplers written in Aesara",
    url="http://github.com/aesara-devs/aemcmc",
    author="aesara-devs",
    author_email="aesara.devs@gmail.com",
    packages=["aemcmc"],
    install_requires=[
        "numpy>=1.18.1",
        "scipy>=1.4.0",
        "aesara>=2.8.11",
        "aeppl>=0.1.2",
        "aehmc>=0.0.10",
        "polyagamma>=1.3.2",
        "cons",
        "logical-unification",
        "etuples",
        "miniKanren",
    ],
    tests_require=["pytest"],
    long_description=open("README.rst").read() if exists("README.rst") else "",
    long_description_content_type="text/x-rst",
    zip_safe=False,
    python_requires=">=3.8",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    keywords=["math", "probability", "numerical", "symbolic", "MCMC"],
)



import aesara
import aesara.tensor as at
import numpy as np
import pytest
from aesara.sparse.basic import as_sparse
from scipy.sparse import csc_matrix

from aemcmc.dists import (
    multivariate_normal_cong2017,
    multivariate_normal_rue2005,
    polyagamma,
)


def test_polyagamma():
    assert polyagamma(size=(5, 4)).eval().shape == (5, 4)

    with pytest.raises(ValueError, match="values of `h` must be positive"):
        polyagamma([1, -0]).eval()

    default_dtype = aesara.config.floatX
    aesara.config.floatX = "float32"
    assert polyagamma().eval().dtype == "float32"
    aesara.config.floatX = default_dtype

    rng1 = at.random.RandomStream(12345)
    res1 = rng1.gen(polyagamma)
    rng2 = at.random.RandomStream(12345)
    res2 = rng2.gen(polyagamma)
    assert res1.eval() == res2.eval()


def test_multivariate_normal_rue2005():
    b = np.array([0, -1, 5])
    var_inv = np.array([1.0, 2.0, 4.0])
    Q = csc_matrix(np.diag(var_inv))

    var = 1.0 / var_inv
    mean = b * var

    srng = at.random.RandomStream(12345)

    def update():
        return multivariate_normal_rue2005(srng, at.as_tensor(b), as_sparse(Q))

    samples_out, updates = aesara.scan(update, n_steps=10000)
    sampling_fn = aesara.function((), samples_out, updates=updates)
    samples = sampling_fn()

    np.testing.assert_allclose(np.mean(samples, axis=0), mean, atol=0.1)
    np.testing.assert_allclose(np.var(samples, axis=0), var, atol=0.1)


def test_multivariate_normal_cong2017():
    nrng = np.random.default_rng(54321)
    X = nrng.standard_normal(size=(50, 100))
    omega, phi = np.linalg.eigh(X @ X.T)
    A = nrng.random(phi.shape[1])
    t = nrng.random(phi.shape[1])

    phi_omega = phi * omega
    cov = np.linalg.inv(A + phi_omega @ phi.T)
    mean = cov @ phi_omega @ t

    srng = at.random.RandomStream(12345)

    def update():
        return multivariate_normal_cong2017(
            srng,
            at.as_tensor(A),
            at.as_tensor(omega),
            at.as_tensor(phi.T),
            at.as_tensor(t),
        )

    samples_out, updates = aesara.scan(update, n_steps=10000)
    sampling_fn = aesara.function((), samples_out, updates=updates)
    samples = sampling_fn()

    np.testing.assert_allclose(np.mean(samples, axis=0), mean, atol=0.05)
    np.testing.assert_allclose(np.var(samples, axis=0, ddof=1), np.diag(cov), atol=0.05)


def test_multivariate_normal_bhattacharya2016_via_cong2017():
    """This test case is to test if the algorithm from Cong et al (2017) can
    be used realiably to implement one described in Bhattacharya (2016) as a
    special case when ``omega`` is the identity matrix.
    """
    nrng = np.random.default_rng(54321)
    X = nrng.standard_normal(size=(50, 100))
    phi = np.linalg.cholesky(X @ X.T).T
    D = nrng.random(phi.shape[1])
    alpha = nrng.random(phi.shape[0])
    precision = phi.T @ phi
    precision[np.diag_indices_from(precision)] += 1.0 / D
    cov = np.linalg.inv(precision)
    mean = cov @ phi.T @ alpha

    srng = at.random.RandomStream(12345)

    def update():
        def multivariate_normal_bhattacharya2016(rng, D, phi, alpha):
            return multivariate_normal_cong2017(
                rng, 1 / D, at.ones(phi.shape[0]), phi, alpha
            )

        return multivariate_normal_bhattacharya2016(
            srng, at.as_tensor(D), at.as_tensor(phi), at.as_tensor(alpha)
        )

    samples_out, updates = aesara.scan(update, n_steps=10000)
    sampling_fn = aesara.function((), samples_out, updates=updates)
    samples = sampling_fn()

    np.testing.assert_allclose(np.mean(samples, axis=0), mean, atol=0.05)
    np.testing.assert_allclose(np.var(samples, axis=0, ddof=1), np.diag(cov), atol=0.05)



import aesara
from aesara.tensor.random import RandomStream

import aemcmc.nuts as nuts


def test_nuts():
    srng = RandomStream(0)
    mu_rv = srng.normal(0, 1, name="mu")
    sigma_rv = srng.halfnormal(0.0, 1.0, name="sigma")
    Y_rv = srng.normal(mu_rv, sigma_rv, name="Y")

    mu_vv = mu_rv.clone()
    sigma_vv = sigma_rv.clone()
    y_vv = Y_rv.clone()

    to_sample_rvs = {mu_rv: mu_vv, sigma_rv: sigma_vv}
    observed = {Y_rv: y_vv}

    sampling_step, updates, (step_size, inverse_mass_matrix) = nuts.step(
        srng, to_sample_rvs, observed
    )

    # Make sure that the state is properly initialized
    sample_steps = [sampling_step[rv] for rv in to_sample_rvs]
    state_fn = aesara.function(
        (
            mu_vv,
            sigma_vv,
            y_vv,
            step_size,
            inverse_mass_matrix,
        ),
        sample_steps,
        updates=updates,
    )
    new_state = state_fn(1.0, 1.0, 1.0, 0.01, [1.0, 1.0])

    # Make sure that the state has advanced
    assert new_state[0] != 1.0
    assert new_state[1] != 1.0



import aesara.tensor as at
from etuples import etuple, etuplize
from kanren import eq, lall
from kanren.facts import Relation, fact
from unification import var

location_scale_family = Relation("location-scale-family")
fact(location_scale_family, at.random.cauchy)
fact(location_scale_family, at.random.gumbel)
fact(location_scale_family, at.random.laplace)
fact(location_scale_family, at.random.logistic)
fact(location_scale_family, at.random.normal)
fact(location_scale_family, at.random.t)


def location_scale_transform(in_expr, out_expr):
    r"""Produce a goal that represents the action of lifting and sinking
    the scale and location parameters of distributions in the location-scale
    family.

    For instance

    .. math::

        \begin{equation*}
           \frac{
                Y \sim \operatorname{P}(0, 1), \quad
                X = \mu + \sigma\,Y
            }{
                X \sim \operatorname{P}\left(\mu, \sigma\right)
            }
        \end{equation*}

    where `P` is any distribution in the location-scale family.

    Parameters
    ----------
    in_expr
        An expression that represents a random variable whose distribution belongs
        to the location-scale family.
    out_expr
        An expression for the non-centered representation of this random variable.

    """

    # Centered representation
    rng_lv, size_lv, type_idx_lv = var(), var(), var()
    mu_lv, sd_lv = var(), var()
    distribution_lv = var()
    centered_et = etuple(distribution_lv, rng_lv, size_lv, type_idx_lv, mu_lv, sd_lv)

    # Non-centered representation
    noncentered_et = etuple(
        etuplize(at.add),
        mu_lv,
        etuple(
            etuplize(at.mul),
            sd_lv,
            etuple(
                distribution_lv,
                0,
                1,
                rng=rng_lv,
                size=size_lv,
                dtype=type_idx_lv,
            ),
        ),
    )

    return lall(
        eq(in_expr, centered_et),
        eq(out_expr, noncentered_et),
        location_scale_family(distribution_lv),
    )


def invgamma_exponential(invgamma_expr, invexponential_expr):
    r"""Produce a goal that represents the relation between the inverse gamma distribution
    and the inverse of an exponential distribution.

    .. math::

        \begin{equation*}
            \frac{
                X \sim \operatorname{Gamma^{-1}}\left(1, c\right),\quad
                Y = 1 / X
            }{
                Y \sim \operatorname{Exp}\left(c\right)
            }
        \end{equation*}

    Parameters
    ----------
    invgamma_expr
        An expression that represents a random variable with an inverse gamma
        distribution with a shape parameter equal to 1.
    invexponential_expr
        An expression that represents the inverse of a random variable with an
        exponential distribution.

    """
    c_lv = var()
    rng_lv, size_lv, dtype_lv = var(), var(), var()

    invgamma_et = etuple(
        etuplize(at.random.invgamma), rng_lv, size_lv, dtype_lv, at.as_tensor(1.0), c_lv
    )

    exponential_et = etuple(
        etuplize(at.random.exponential),
        c_lv,
        rng=rng_lv,
        size=size_lv,
        dtype=dtype_lv,
    )
    invexponential_et = etuple(at.true_divide, at.as_tensor(1.0), exponential_et)

    return lall(
        eq(invgamma_expr, invgamma_et), eq(invexponential_expr, invexponential_et)
    )
